"""
RDP PDU layer implementation.

This module implements RDP-specific PDUs for the connection sequence
and data exchange phases.
"""

import struct
from logging import getLogger
from typing import Any

logger = getLogger(__name__)

# Security header flags
SEC_EXCHANGE_PKT = 0x0001
SEC_TRANSPORT_REQ = 0x0002
SEC_TRANSPORT_RSP = 0x0004
SEC_ENCRYPT = 0x0008
SEC_RESET_SEQNO = 0x0010
SEC_IGNORE_SEQNO = 0x0020
SEC_INFO_PKT = 0x0040
SEC_LICENSE_PKT = 0x0080
SEC_LICENSE_ENCRYPT_CS = 0x0200
SEC_LICENSE_ENCRYPT_SC = 0x0200
SEC_REDIRECTION_PKT = 0x0400
SEC_SECURE_CHECKSUM = 0x0800
SEC_AUTODETECT_REQ = 0x1000
SEC_AUTODETECT_RSP = 0x2000
SEC_HEARTBEAT = 0x4000
SEC_FLAGSHI_VALID = 0x8000

# Share control PDU types
PDUTYPE_DEMANDACTIVEPDU = 0x0001
PDUTYPE_CONFIRMACTIVEPDU = 0x0003
PDUTYPE_DEACTIVATEALLPDU = 0x0006
PDUTYPE_DATAPDU = 0x0007
PDUTYPE_SERVER_REDIR_PKT = 0x000A

# Share data PDU types
PDUTYPE2_UPDATE = 0x02
PDUTYPE2_CONTROL = 0x14
PDUTYPE2_POINTER = 0x1B
PDUTYPE2_INPUT = 0x1C
PDUTYPE2_SYNCHRONIZE = 0x1F
PDUTYPE2_REFRESH_RECT = 0x21
PDUTYPE2_PLAY_SOUND = 0x22
PDUTYPE2_SUPPRESS_OUTPUT = 0x23
PDUTYPE2_SHUTDOWN_REQUEST = 0x24
PDUTYPE2_SHUTDOWN_DENIED = 0x25
PDUTYPE2_SAVE_SESSION_INFO = 0x26
PDUTYPE2_FONTLIST = 0x27
PDUTYPE2_FONTMAP = 0x28
PDUTYPE2_SET_KEYBOARD_INDICATORS = 0x29
PDUTYPE2_BITMAPCACHE_PERSISTENT_LIST = 0x2B
PDUTYPE2_BITMAPCACHE_ERROR_PDU = 0x2C
PDUTYPE2_SET_KEYBOARD_IME_STATUS = 0x2D
PDUTYPE2_OFFSCRCACHE_ERROR_PDU = 0x2E
PDUTYPE2_SET_ERROR_INFO_PDU = 0x2F
PDUTYPE2_DRAWNINEGRID_ERROR_PDU = 0x30
PDUTYPE2_DRAWGDIPLUS_ERROR_PDU = 0x31
PDUTYPE2_ARC_STATUS_PDU = 0x32
PDUTYPE2_STATUS_INFO_PDU = 0x36
PDUTYPE2_MONITOR_LAYOUT_PDU = 0x37

# Control actions
CTRLACTION_REQUEST_CONTROL = 0x0001
CTRLACTION_GRANTED_CONTROL = 0x0002
CTRLACTION_DETACH = 0x0003
CTRLACTION_COOPERATE = 0x0004

# Input event types
INPUT_EVENT_SYNC = 0x0000
INPUT_EVENT_UNUSED = 0x0002
INPUT_EVENT_SCANCODE = 0x0004
INPUT_EVENT_UNICODE = 0x0005
INPUT_EVENT_MOUSE = 0x8001
INPUT_EVENT_MOUSEX = 0x8002

# Mouse event flags
PTRFLAGS_HWHEEL = 0x0400
PTRFLAGS_WHEEL = 0x0200
PTRFLAGS_WHEEL_NEGATIVE = 0x0100
PTRFLAGS_MOVE = 0x0800
PTRFLAGS_DOWN = 0x8000
PTRFLAGS_BUTTON1 = 0x1000  # Left button
PTRFLAGS_BUTTON2 = 0x2000  # Right button
PTRFLAGS_BUTTON3 = 0x4000  # Middle button

# Keyboard event flags
KBDFLAGS_EXTENDED = 0x0100
KBDFLAGS_EXTENDED1 = 0x0200
KBDFLAGS_DOWN = 0x4000
KBDFLAGS_RELEASE = 0x8000

# Update types
UPDATETYPE_ORDERS = 0x0000
UPDATETYPE_BITMAP = 0x0001
UPDATETYPE_PALETTE = 0x0002
UPDATETYPE_SYNCHRONIZE = 0x0003

# Info packet flags
INFO_MOUSE = 0x00000001
INFO_DISABLECTRLALTDEL = 0x00000002
INFO_AUTOLOGON = 0x00000008
INFO_UNICODE = 0x00000010
INFO_MAXIMIZESHELL = 0x00000020
INFO_LOGONNOTIFY = 0x00000040
INFO_COMPRESSION = 0x00000080
INFO_ENABLEWINDOWSKEY = 0x00000100
INFO_REMOTECONSOLEAUDIO = 0x00002000
INFO_FORCE_ENCRYPTED_CS_PDU = 0x00004000
INFO_RAIL = 0x00008000
INFO_LOGONERRORS = 0x00010000
INFO_MOUSE_HAS_WHEEL = 0x00020000
INFO_PASSWORD_IS_SC_PIN = 0x00040000
INFO_NOAUDIOPLAYBACK = 0x00080000
INFO_USING_SAVED_CREDS = 0x00100000
INFO_AUDIOCAPTURE = 0x00200000
INFO_VIDEO_DISABLE = 0x00400000
INFO_HIDEF_RAIL_SUPPORTED = 0x02000000

# Performance flags
PERF_DISABLE_WALLPAPER = 0x00000001
PERF_DISABLE_FULLWINDOWDRAG = 0x00000002
PERF_DISABLE_MENUANIMATIONS = 0x00000004
PERF_DISABLE_THEMING = 0x00000008
PERF_DISABLE_CURSOR_SHADOW = 0x00000020
PERF_DISABLE_CURSORSETTINGS = 0x00000040
PERF_ENABLE_FONT_SMOOTHING = 0x00000080
PERF_ENABLE_DESKTOP_COMPOSITION = 0x00000100


def build_client_info_pdu(
    domain: str = "",
    username: str = "",
    password: str = "",
    shell: str = "",
    work_dir: str = "",
    flags: int = INFO_MOUSE | INFO_UNICODE | INFO_LOGONNOTIFY | INFO_DISABLECTRLALTDEL,
    performance_flags: int = PERF_DISABLE_WALLPAPER,
) -> bytes:
    """
    Build Client Info PDU (TS_INFO_PACKET).

    This is sent after the channel join sequence to provide user credentials
    and session configuration.
    """
    data = bytearray()

    # Code page (4 bytes) - not used with INFO_UNICODE
    data += struct.pack("<I", 0)

    # Flags (4 bytes)
    data += struct.pack("<I", flags)

    # Encode strings as UTF-16LE (without null terminator for length fields)
    domain_bytes = domain.encode("utf-16-le")
    username_bytes = username.encode("utf-16-le")
    password_bytes = password.encode("utf-16-le")
    shell_bytes = shell.encode("utf-16-le")
    work_dir_bytes = work_dir.encode("utf-16-le")

    # String lengths (in bytes, excluding null terminator)
    data += struct.pack("<H", len(domain_bytes))
    data += struct.pack("<H", len(username_bytes))
    data += struct.pack("<H", len(password_bytes))
    data += struct.pack("<H", len(shell_bytes))
    data += struct.pack("<H", len(work_dir_bytes))

    # Strings with null terminators
    data += domain_bytes + b"\x00\x00"
    data += username_bytes + b"\x00\x00"
    data += password_bytes + b"\x00\x00"
    data += shell_bytes + b"\x00\x00"
    data += work_dir_bytes + b"\x00\x00"

    # Extended Info (TS_EXTENDED_INFO_PACKET)
    # Client address family (2 bytes) - AF_INET = 0x0002
    data += struct.pack("<H", 0x0002)

    # Client address length (2 bytes)
    client_address = b"\x00\x00"  # Empty address
    data += struct.pack("<H", len(client_address))
    data += client_address

    # Client directory length (2 bytes)
    client_dir = b"\x00\x00"
    data += struct.pack("<H", len(client_dir))
    data += client_dir

    # Time zone info (172 bytes) - simplified, all zeros
    data += bytes(172)

    # Client session ID (4 bytes)
    data += struct.pack("<I", 0)

    # Performance flags (4 bytes)
    data += struct.pack("<I", performance_flags)

    # Auto reconnect cookie length (2 bytes)
    data += struct.pack("<H", 0)

    return bytes(data)


def build_security_exchange_pdu(encrypted_client_random: bytes) -> bytes:
    """
    Build Security Exchange PDU.

    This is only used with Standard RDP Security (not TLS/NLA).
    """
    data = bytearray()

    # Security header flags (4 bytes)
    data += struct.pack("<I", SEC_EXCHANGE_PKT)

    # Length of encrypted client random (4 bytes)
    data += struct.pack("<I", len(encrypted_client_random) + 8)

    # Encrypted client random
    data += encrypted_client_random

    # Padding (8 bytes)
    data += bytes(8)

    return bytes(data)


def build_share_control_header(pdu_type: int, pdu_source: int, share_id: int = 0) -> bytes:
    """Build Share Control Header."""
    data = bytearray()

    # Total length placeholder (2 bytes) - will be filled by caller
    data += struct.pack("<H", 0)

    # PDU type (2 bytes)
    data += struct.pack("<H", pdu_type | 0x0010)  # Version 1

    # PDU source (2 bytes)
    data += struct.pack("<H", pdu_source)

    return bytes(data)


def build_share_data_header(
    share_id: int,
    pdu_source: int,
    pdu_type2: int,
    compressed_type: int = 0,
    compressed_length: int = 0,
) -> bytes:
    """Build Share Data Header."""
    data = bytearray()

    # Share ID (4 bytes)
    data += struct.pack("<I", share_id)

    # Padding (1 byte)
    data += bytes([0])

    # Stream ID (1 byte) - STREAM_MED = 0x01
    data += bytes([0x01])

    # Uncompressed length placeholder (2 bytes)
    data += struct.pack("<H", 0)

    # PDU type 2 (1 byte)
    data += bytes([pdu_type2])

    # Compressed type (1 byte)
    data += bytes([compressed_type])

    # Compressed length (2 bytes)
    data += struct.pack("<H", compressed_length)

    return bytes(data)


def build_synchronize_pdu(target_user: int) -> bytes:
    """Build Synchronize PDU."""
    data = bytearray()

    # Message type (2 bytes) - SYNCMSGTYPE_SYNC = 0x0001
    data += struct.pack("<H", 0x0001)

    # Target user (2 bytes)
    data += struct.pack("<H", target_user)

    return bytes(data)


def build_control_pdu(action: int, grant_id: int = 0, control_id: int = 0) -> bytes:
    """Build Control PDU."""
    data = bytearray()

    # Action (2 bytes)
    data += struct.pack("<H", action)

    # Grant ID (2 bytes)
    data += struct.pack("<H", grant_id)

    # Control ID (4 bytes)
    data += struct.pack("<I", control_id)

    return bytes(data)


def build_font_list_pdu() -> bytes:
    """Build Font List PDU."""
    data = bytearray()

    # Number of fonts (2 bytes)
    data += struct.pack("<H", 0)

    # Total number of fonts (2 bytes)
    data += struct.pack("<H", 0)

    # List flags (2 bytes) - FONTLIST_FIRST | FONTLIST_LAST = 0x0003
    data += struct.pack("<H", 0x0003)

    # Entry size (2 bytes)
    data += struct.pack("<H", 0x0032)

    return bytes(data)


def build_input_event_pdu(events: list[tuple[int, int, bytes]]) -> bytes:
    """
    Build Input Event PDU.

    Args:
        events: List of (event_time, event_type, event_data) tuples
    """
    data = bytearray()

    # Number of events (2 bytes)
    data += struct.pack("<H", len(events))

    # Padding (2 bytes)
    data += struct.pack("<H", 0)

    # Events
    for event_time, event_type, event_data in events:
        # Event time (4 bytes)
        data += struct.pack("<I", event_time)
        # Event type (2 bytes)
        data += struct.pack("<H", event_type)
        # Event data
        data += event_data

    return bytes(data)


def build_scancode_event(
    scan_code: int,
    is_release: bool = False,
    is_extended: bool = False,
) -> bytes:
    """Build keyboard scancode event data."""
    flags = 0
    if is_release:
        flags |= KBDFLAGS_RELEASE
    if is_extended:
        flags |= KBDFLAGS_EXTENDED

    data = bytearray()
    # Key flags (2 bytes)
    data += struct.pack("<H", flags)
    # Key code (2 bytes)
    data += struct.pack("<H", scan_code)
    # Padding (2 bytes)
    data += struct.pack("<H", 0)

    return bytes(data)


def build_unicode_event(unicode_code: int, is_release: bool = False) -> bytes:
    """Build keyboard unicode event data."""
    flags = 0
    if is_release:
        flags |= KBDFLAGS_RELEASE

    data = bytearray()
    # Key flags (2 bytes)
    data += struct.pack("<H", flags)
    # Unicode code (2 bytes)
    data += struct.pack("<H", unicode_code)
    # Padding (2 bytes)
    data += struct.pack("<H", 0)

    return bytes(data)


def build_mouse_event(
    x: int,
    y: int,
    button: int = 0,  # 0=none, 1=left, 2=right, 3=middle
    is_down: bool = False,
    is_move: bool = True,
    wheel_delta: int = 0,  # Positive=up, negative=down
) -> bytes:
    """Build mouse event data."""
    flags = 0

    if is_move:
        flags |= PTRFLAGS_MOVE

    if button == 1:
        flags |= PTRFLAGS_BUTTON1
    elif button == 2:
        flags |= PTRFLAGS_BUTTON2
    elif button == 3:
        flags |= PTRFLAGS_BUTTON3

    if is_down and button > 0:
        flags |= PTRFLAGS_DOWN

    # Handle mouse wheel
    if wheel_delta != 0:
        flags |= PTRFLAGS_WHEEL
        if wheel_delta < 0:
            flags |= PTRFLAGS_WHEEL_NEGATIVE
            wheel_delta = -wheel_delta
        # Wheel delta is in the lower 9 bits (0-511), typically 120 per notch
        flags |= wheel_delta & 0x01FF

    data = bytearray()
    # Pointer flags (2 bytes)
    data += struct.pack("<H", flags)
    # X position (2 bytes)
    data += struct.pack("<H", x & 0xFFFF)
    # Y position (2 bytes)
    data += struct.pack("<H", y & 0xFFFF)

    return bytes(data)


def build_confirm_active_pdu(
    share_id: int,
    originator_id: int,
    source_descriptor: bytes,
    capabilities: bytes,
) -> bytes:
    """Build Confirm Active PDU."""
    data = bytearray()

    # Share ID (4 bytes)
    data += struct.pack("<I", share_id)

    # Originator ID (2 bytes)
    data += struct.pack("<H", originator_id)

    # Source descriptor length (2 bytes)
    data += struct.pack("<H", len(source_descriptor))

    # Combined capabilities length (2 bytes)
    data += struct.pack("<H", len(capabilities))

    # Source descriptor
    data += source_descriptor

    # Number of capability sets (2 bytes) - encoded in capabilities
    # Padding (2 bytes)
    # Capability sets

    data += capabilities

    return bytes(data)


def parse_demand_active_pdu(data: bytes) -> dict[str, Any]:
    """Parse Demand Active PDU from server."""
    result: dict[str, Any] = {
        "share_id": 0,
        "source_descriptor": b"",
        "capabilities": [],
    }

    offset = 0

    # Share ID (4 bytes)
    result["share_id"] = struct.unpack_from("<I", data, offset)[0]
    offset += 4

    # Length of source descriptor (2 bytes)
    source_len = struct.unpack_from("<H", data, offset)[0]
    offset += 2

    # Combined capabilities length (2 bytes)
    _caps_len = struct.unpack_from("<H", data, offset)[0]  # noqa: F841
    offset += 2

    # Source descriptor
    result["source_descriptor"] = data[offset : offset + source_len]
    offset += source_len

    # Number of capability sets (2 bytes)
    num_caps = struct.unpack_from("<H", data, offset)[0]
    offset += 2

    # Padding (2 bytes)
    offset += 2

    # Parse capability sets
    for _ in range(num_caps):
        if offset + 4 > len(data):
            break

        cap_type = struct.unpack_from("<H", data, offset)[0]
        cap_len = struct.unpack_from("<H", data, offset + 2)[0]

        if cap_len < 4:
            break

        cap_data = data[offset : offset + cap_len]
        result["capabilities"].append({"type": cap_type, "data": cap_data})
        offset += cap_len

    return result


def parse_update_pdu(data: bytes) -> dict[str, Any]:
    """Parse Update PDU from server."""
    result: dict[str, Any] = {"update_type": 0, "data": b""}

    if len(data) < 2:
        return result

    result["update_type"] = struct.unpack_from("<H", data, 0)[0]
    result["data"] = data[2:]

    return result


def parse_bitmap_update(data: bytes) -> list[dict[str, Any]]:
    """Parse Bitmap Update data."""
    bitmaps: list[dict[str, Any]] = []

    if len(data) < 2:
        return bitmaps

    offset = 0
    num_rects = struct.unpack_from("<H", data, offset)[0]
    offset += 2

    logger.debug(f"Parsing bitmap update: {len(data)} bytes, {num_rects} rectangles")

    for i in range(num_rects):
        if offset + 18 > len(data):
            logger.debug(f"Stopping at rect {i}: not enough data for header (offset={offset}, data_len={len(data)})")
            break

        bitmap = {
            "dest_left": struct.unpack_from("<H", data, offset)[0],
            "dest_top": struct.unpack_from("<H", data, offset + 2)[0],
            "dest_right": struct.unpack_from("<H", data, offset + 4)[0],
            "dest_bottom": struct.unpack_from("<H", data, offset + 6)[0],
            "width": struct.unpack_from("<H", data, offset + 8)[0],
            "height": struct.unpack_from("<H", data, offset + 10)[0],
            "bpp": struct.unpack_from("<H", data, offset + 12)[0],
            "flags": struct.unpack_from("<H", data, offset + 14)[0],
            "length": struct.unpack_from("<H", data, offset + 16)[0],
        }
        offset += 18

        # Validate bitmap data
        if bitmap["bpp"] not in (8, 15, 16, 24, 32):
            logger.debug(f"Rect {i}: Invalid bpp={bitmap['bpp']}, stopping parse (likely corrupt data)")
            break

        if bitmap["width"] == 0 or bitmap["height"] == 0:
            logger.debug(f"Rect {i}: Invalid dimensions {bitmap['width']}x{bitmap['height']}, skipping")
            offset += bitmap["length"]
            continue

        if offset + bitmap["length"] > len(data):
            logger.debug(
                f"Stopping at rect {i}: not enough data for bitmap (need {bitmap['length']}, have {len(data) - offset})"
            )
            break

        bitmap["data"] = data[offset : offset + bitmap["length"]]
        offset += bitmap["length"]

        bitmaps.append(bitmap)

    logger.debug(f"Parsed {len(bitmaps)} valid bitmaps")
    return bitmaps


def build_refresh_rect_pdu(rectangles: list[tuple[int, int, int, int]]) -> bytes:
    """
    Build Refresh Rect PDU to request screen redraw.

    This PDU is sent by the client to request the server to redraw
    one or more rectangles of the session screen area.

    Args:
        rectangles: List of (left, top, right, bottom) tuples defining
                   the areas to refresh.

    Returns:
        The Refresh Rect PDU data (without Share Data Header).
    """
    data = bytearray()

    # numberOfAreas (1 byte)
    data += bytes([len(rectangles)])

    # pad3Octets (3 bytes)
    data += bytes(3)

    # areasToRefresh - array of TS_RECTANGLE16
    for left, top, right, bottom in rectangles:
        data += struct.pack("<H", left)
        data += struct.pack("<H", top)
        data += struct.pack("<H", right)
        data += struct.pack("<H", bottom)

    return bytes(data)


def build_suppress_output_pdu(allow_display_updates: bool, rectangle: tuple[int, int, int, int] | None = None) -> bytes:
    """
    Build Suppress Output PDU to control display updates.

    This PDU is sent by the client to toggle the sending of
    desktop display updates from the server.

    Args:
        allow_display_updates: If True, the server should send display updates.
                              If False, the server should stop sending updates.
        rectangle: Optional (left, top, right, bottom) tuple defining the
                  desktop rectangle to allow updates for (only used when
                  allow_display_updates is True).

    Returns:
        The Suppress Output PDU data (without Share Data Header).
    """
    data = bytearray()

    # allowDisplayUpdates (1 byte) - 0x00 = suppress, 0x01 = allow
    data += bytes([0x01 if allow_display_updates else 0x00])

    # pad3Octets (3 bytes)
    data += bytes(3)

    # desktopRect (TS_RECTANGLE16) - only if allowDisplayUpdates is TRUE
    if allow_display_updates and rectangle:
        left, top, right, bottom = rectangle
        data += struct.pack("<H", left)
        data += struct.pack("<H", top)
        data += struct.pack("<H", right)
        data += struct.pack("<H", bottom)

    return bytes(data)



"""Tests for Screen and Display classes."""

from simple_rdp.display import Display
from simple_rdp.display import ScreenBuffer


class TestScreenBuffer:
    """Tests for ScreenBuffer dataclass."""

    def test_screen_buffer_creation(self):
        """Test ScreenBuffer can be created with required fields."""
        buffer = ScreenBuffer(width=1920, height=1080, data=b"\x00" * 100)
        assert buffer.width == 1920
        assert buffer.height == 1080
        assert buffer.format == "RGB"

    def test_screen_buffer_custom_format(self):
        """Test ScreenBuffer with custom format."""
        buffer = ScreenBuffer(width=800, height=600, data=b"\x00", format="RGBA")
        assert buffer.format == "RGBA"


class TestDisplay:
    """Tests for Display class."""

    def test_initial_state(self):
        """Test initial state of Display."""
        display = Display(width=1920, height=1080)
        assert display.width == 1920
        assert display.height == 1080
        assert display.frame_count == 0

    def test_display_stats(self):
        """Test display statistics."""
        display = Display(width=1920, height=1080, fps=30)
        stats = display.stats
        assert stats["frames_received"] == 0
        assert stats["frames_encoded"] == 0
        assert stats["encoding_errors"] == 0



"""
Video recording of RDP session with automated interactions.

This example demonstrates:
- Native file recording using library's built-in ffmpeg streaming
- Automated mouse and keyboard interactions
- Pointer composited into video frames
"""

import asyncio
import logging
import os
from datetime import datetime
from pathlib import Path

from dotenv import load_dotenv
from rich.logging import RichHandler

from simple_rdp import RDPClient

load_dotenv()


async def perform_interactions(client: RDPClient) -> None:
    """
    Perform automated mouse and keyboard interactions.

    This runs for approximately 1 minute, performing various UI interactions.
    """
    print("\n=== Starting automated interactions ===\n")

    # Screen dimensions (assuming 1920x1080)
    width, height = 1920, 1080

    # Windows button location (bottom-left, Start button)
    start_button = (28, height - 28)  # ~center of Start button

    # Clock/system tray location (bottom-right)
    clock_area = (width - 100, height - 28)

    # Center of screen
    center = (width // 2, height // 2)

    interactions = [
        # Phase 1: Initial setup (0-10 seconds)
        ("Move mouse to center", lambda: client.mouse_move(*center)),
        ("Wait", lambda: asyncio.sleep(1)),
        ("Click center", lambda: client.mouse_click(*center)),
        ("Wait", lambda: asyncio.sleep(2)),
        # Phase 2: Windows Start menu (10-20 seconds)
        ("Move to Start button", lambda: client.mouse_move(*start_button)),
        ("Wait", lambda: asyncio.sleep(0.5)),
        ("Click Start button", lambda: client.mouse_click(*start_button)),
        ("Wait for Start menu", lambda: asyncio.sleep(2)),
        ("Press Windows key to toggle", lambda: client.send_key(0x5B)),
        ("Wait", lambda: asyncio.sleep(1)),
        # Phase 3: Open Start menu with keyboard (20-30 seconds)
        ("Press Windows key", lambda: client.send_key(0x5B)),
        ("Wait for menu", lambda: asyncio.sleep(1.5)),
        ("Type 'settings'", lambda: client.send_text("settings")),
        ("Wait for search", lambda: asyncio.sleep(2)),
        ("Press Escape", lambda: client.send_key(0x01)),
        ("Wait", lambda: asyncio.sleep(1)),
        # Phase 4: Mouse movements (30-40 seconds)
        ("Move mouse top-left", lambda: client.mouse_move(100, 100)),
        ("Wait", lambda: asyncio.sleep(0.5)),
        ("Move mouse top-right", lambda: client.mouse_move(width - 100, 100)),
        ("Wait", lambda: asyncio.sleep(0.5)),
        ("Move mouse bottom-right", lambda: client.mouse_move(width - 100, height - 100)),
        ("Wait", lambda: asyncio.sleep(0.5)),
        ("Move mouse bottom-left", lambda: client.mouse_move(100, height - 100)),
        ("Wait", lambda: asyncio.sleep(0.5)),
        ("Move mouse to center", lambda: client.mouse_move(*center)),
        ("Wait", lambda: asyncio.sleep(1)),
        # Phase 5: Click on clock/system tray (40-50 seconds)
        ("Move to clock area", lambda: client.mouse_move(*clock_area)),
        ("Wait", lambda: asyncio.sleep(0.5)),
        ("Click clock", lambda: client.mouse_click(*clock_area)),
        ("Wait for calendar popup", lambda: asyncio.sleep(3)),
        ("Press Escape to close", lambda: client.send_key(0x01)),
        ("Wait", lambda: asyncio.sleep(1)),
        # Phase 6: Final interactions (50-60 seconds)
        ("Press Windows key", lambda: client.send_key(0x5B)),
        ("Wait", lambda: asyncio.sleep(1.5)),
        ("Type 'notepad'", lambda: client.send_text("notepad")),
        ("Wait", lambda: asyncio.sleep(2)),
        ("Press Escape", lambda: client.send_key(0x01)),
        ("Wait", lambda: asyncio.sleep(1)),
        ("Move to center", lambda: client.mouse_move(*center)),
        ("Final wait", lambda: asyncio.sleep(2)),
    ]

    for i, (description, action) in enumerate(interactions):
        print(f"  [{i + 1}/{len(interactions)}] {description}")
        await action()


async def main() -> None:
    logging.basicConfig(level=logging.INFO, format="%(message)s", handlers=[RichHandler()])

    # Get RDP connection details from environment
    host = os.environ.get("RDP_HOST", "")
    username = os.environ.get("RDP_USER", "")
    password = os.environ.get("RDP_PASS", "")

    if not host or not username or not password:
        print("ERROR: Set RDP_HOST, RDP_USER, RDP_PASS environment variables (or use .env file)")
        return

    # Create session directory with timestamp
    timestamp = datetime.now().strftime("%Y%m%d_%H%M%S")
    session_dir = Path("sessions") / timestamp
    session_dir.mkdir(parents=True, exist_ok=True)
    video_path = session_dir / "recording.ts"

    print(f"\n{'=' * 60}")
    print(f"RDP Session Recording - {timestamp}")
    print(f"Session directory: {session_dir}")
    print(f"Video output: {video_path}")
    print(f"{'=' * 60}\n")

    async with RDPClient(host=host, username=username, password=password, show_wallpaper=True) as client:
        print("Connected to RDP server!\n")

        # Wait a moment for initial screen to render
        await asyncio.sleep(2)

        # Start native file recording (uses library's built-in ffmpeg streaming)
        print("Starting native file recording...")
        await client.start_file_recording(str(video_path))
        print(f"Recording to: {video_path}")

        try:
            # Perform automated interactions (~1 minute)
            await perform_interactions(client)

        finally:
            # Stop recording
            print("\nStopping recording...")
            await client.stop_file_recording()
            await client.stop_streaming()

    # Show results
    print(f"\n{'=' * 60}")
    print("Recording complete!")
    print(f"{'=' * 60}\n")

    if video_path.exists():
        size_mb = video_path.stat().st_size / (1024 * 1024)
        stats = client.get_recording_stats()
        print(f"✓ Video saved: {video_path}")
        print(f"  Size: {size_mb:.2f} MB")
        print(f"  Frames encoded: {stats.get('frames_encoded', 'N/A')}")
        print(f"  Pointer updates: {stats.get('pointer_updates', 'N/A')}")
        print(f"  Pointer throttled: {stats.get('pointer_updates_throttled', 'N/A')}")
    else:
        print("✗ Video file not found")

    print(f"\n{'=' * 60}")
    print("Session complete!")
    print(f"{'=' * 60}\n")


if __name__ == "__main__":
    asyncio.run(main())



"""Tests for CredSSP module."""

import pytest

from simple_rdp.credssp import ASN1_CONTEXT_0
from simple_rdp.credssp import ASN1_INTEGER
from simple_rdp.credssp import ASN1_OCTET_STRING
from simple_rdp.credssp import ASN1_SEQUENCE
from simple_rdp.credssp import CREDSSP_VERSION
from simple_rdp.credssp import NONCE_SIZE
from simple_rdp.credssp import _decode_asn1_element
from simple_rdp.credssp import _decode_asn1_length
from simple_rdp.credssp import _encode_asn1_context
from simple_rdp.credssp import _encode_asn1_integer
from simple_rdp.credssp import _encode_asn1_length
from simple_rdp.credssp import _encode_asn1_octet_string
from simple_rdp.credssp import _encode_asn1_sequence
from simple_rdp.credssp import build_ts_credentials
from simple_rdp.credssp import build_ts_request
from simple_rdp.credssp import build_ts_request_with_credentials
from simple_rdp.credssp import build_ts_request_with_pub_key_auth
from simple_rdp.credssp import parse_ts_request


class TestConstants:
    """Tests for CredSSP constants."""

    def test_credssp_version(self) -> None:
        """Test CredSSP version constant."""
        assert CREDSSP_VERSION == 6

    def test_nonce_size(self) -> None:
        """Test nonce size constant."""
        assert NONCE_SIZE == 32

    def test_asn1_tags(self) -> None:
        """Test ASN.1 tag constants."""
        assert ASN1_SEQUENCE == 0x30
        assert ASN1_CONTEXT_0 == 0xA0
        assert ASN1_INTEGER == 0x02
        assert ASN1_OCTET_STRING == 0x04


class TestAsn1LengthEncoding:
    """Tests for ASN.1 length encoding."""

    def test_encode_short_length(self) -> None:
        """Test encoding short lengths (< 128)."""
        assert _encode_asn1_length(0) == bytes([0])
        assert _encode_asn1_length(127) == bytes([127])
        assert _encode_asn1_length(50) == bytes([50])

    def test_encode_medium_length(self) -> None:
        """Test encoding medium lengths (128-255)."""
        assert _encode_asn1_length(128) == bytes([0x81, 128])
        assert _encode_asn1_length(255) == bytes([0x81, 255])

    def test_encode_long_length(self) -> None:
        """Test encoding long lengths (256-65535)."""
        result = _encode_asn1_length(256)
        assert result == bytes([0x82, 0x01, 0x00])
        result = _encode_asn1_length(1000)
        assert result == bytes([0x82, 0x03, 0xE8])


class TestAsn1LengthDecoding:
    """Tests for ASN.1 length decoding."""

    def test_decode_short_length(self) -> None:
        """Test decoding short lengths."""
        length, consumed = _decode_asn1_length(bytes([50, 0x00]), 0)
        assert length == 50
        assert consumed == 1

    def test_decode_long_form_1_byte(self) -> None:
        """Test decoding long form 1 byte length."""
        length, consumed = _decode_asn1_length(bytes([0x81, 200]), 0)
        assert length == 200
        assert consumed == 2

    def test_decode_long_form_2_bytes(self) -> None:
        """Test decoding long form 2 byte length."""
        length, consumed = _decode_asn1_length(bytes([0x82, 0x01, 0x00]), 0)
        assert length == 256
        assert consumed == 3


class TestAsn1IntegerEncoding:
    """Tests for ASN.1 integer encoding."""

    def test_encode_small_integer(self) -> None:
        """Test encoding small integers."""
        result = _encode_asn1_integer(0)
        assert result[0] == ASN1_INTEGER
        assert result[2] == 0

    def test_encode_medium_integer(self) -> None:
        """Test encoding medium integers."""
        result = _encode_asn1_integer(100)
        assert result[0] == ASN1_INTEGER

    def test_encode_large_integer(self) -> None:
        """Test encoding large integers."""
        result = _encode_asn1_integer(65535)
        assert result[0] == ASN1_INTEGER
        assert len(result) > 3


class TestAsn1OctetString:
    """Tests for ASN.1 octet string encoding."""

    def test_encode_empty_octet_string(self) -> None:
        """Test encoding empty octet string."""
        result = _encode_asn1_octet_string(b"")
        assert result == bytes([ASN1_OCTET_STRING, 0])

    def test_encode_octet_string_with_data(self) -> None:
        """Test encoding octet string with data."""
        data = b"hello"
        result = _encode_asn1_octet_string(data)
        assert result[0] == ASN1_OCTET_STRING
        assert result[1] == len(data)
        assert result[2:] == data


class TestAsn1Context:
    """Tests for ASN.1 context-specific encoding."""

    def test_encode_context_0(self) -> None:
        """Test encoding with context tag 0."""
        content = b"\x01\x02"
        result = _encode_asn1_context(ASN1_CONTEXT_0, content)
        assert result[0] == ASN1_CONTEXT_0
        assert result[1] == len(content)
        assert result[2:] == content


class TestAsn1Sequence:
    """Tests for ASN.1 sequence encoding."""

    def test_encode_empty_sequence(self) -> None:
        """Test encoding empty sequence."""
        result = _encode_asn1_sequence(b"")
        assert result == bytes([ASN1_SEQUENCE, 0])

    def test_encode_sequence_with_content(self) -> None:
        """Test encoding sequence with content."""
        content = b"\x02\x01\x05"  # INTEGER 5
        result = _encode_asn1_sequence(content)
        assert result[0] == ASN1_SEQUENCE
        assert result[1] == len(content)
        assert result[2:] == content


class TestAsn1ElementDecoding:
    """Tests for ASN.1 element decoding."""

    def test_decode_integer_element(self) -> None:
        """Test decoding an integer element."""
        data = bytes([ASN1_INTEGER, 0x01, 0x05])
        tag, content, consumed = _decode_asn1_element(data, 0)
        assert tag == ASN1_INTEGER
        assert content == bytes([0x05])
        assert consumed == 3

    def test_decode_sequence_element(self) -> None:
        """Test decoding a sequence element."""
        inner = bytes([ASN1_INTEGER, 0x01, 0x05])
        data = bytes([ASN1_SEQUENCE, len(inner)]) + inner
        tag, content, consumed = _decode_asn1_element(data, 0)
        assert tag == ASN1_SEQUENCE
        assert content == inner
        assert consumed == len(data)


class TestTsRequest:
    """Tests for TSRequest building."""

    def test_build_ts_request_basic(self) -> None:
        """Test building basic TSRequest."""
        result = build_ts_request()
        assert isinstance(result, bytes)
        assert len(result) > 0
        assert result[0] == ASN1_SEQUENCE

    def test_build_ts_request_with_nego_token(self) -> None:
        """Test building TSRequest with nego token."""
        nego_token = b"test_token_data"
        result = build_ts_request(nego_token=nego_token)
        assert isinstance(result, bytes)
        assert len(result) > len(nego_token)

    def test_build_ts_request_custom_version(self) -> None:
        """Test building TSRequest with custom version."""
        result = build_ts_request(version=3)
        assert isinstance(result, bytes)


class TestTsRequestWithPubKeyAuth:
    """Tests for TSRequest with pubKeyAuth."""

    def test_build_ts_request_with_pub_key_auth(self) -> None:
        """Test building TSRequest with pubKeyAuth."""
        pub_key_auth = b"encrypted_pub_key"
        result = build_ts_request_with_pub_key_auth(pub_key_auth)
        assert isinstance(result, bytes)
        assert len(result) > 0

    def test_build_ts_request_with_client_nonce(self) -> None:
        """Test building TSRequest with client nonce."""
        pub_key_auth = b"encrypted_pub_key"
        client_nonce = b"a" * NONCE_SIZE
        result = build_ts_request_with_pub_key_auth(
            pub_key_auth=pub_key_auth,
            client_nonce=client_nonce,
        )
        assert isinstance(result, bytes)


class TestTsRequestWithCredentials:
    """Tests for TSRequest with credentials."""

    def test_build_ts_request_with_credentials(self) -> None:
        """Test building TSRequest with credentials."""
        auth_info = b"encrypted_credentials"
        result = build_ts_request_with_credentials(auth_info)
        assert isinstance(result, bytes)
        assert len(result) > 0


class TestTsCredentials:
    """Tests for TSCredentials building."""

    def test_build_ts_credentials(self) -> None:
        """Test building TSCredentials."""
        # Build credentials with domain, user, password
        result = build_ts_credentials(
            domain="DOMAIN",
            username="user",
            password="pass",
        )
        assert isinstance(result, bytes)
        assert len(result) > 0


class TestParseTsRequest:
    """Tests for TSRequest parsing."""

    def test_parse_ts_request_basic(self) -> None:
        """Test parsing basic TSRequest."""
        # Build a request and parse it back
        original = build_ts_request()
        result = parse_ts_request(original)
        assert "version" in result
        assert result["version"] == CREDSSP_VERSION

    def test_parse_ts_request_with_nego_token(self) -> None:
        """Test parsing TSRequest with nego token."""
        nego_token = b"test_token"
        original = build_ts_request(nego_token=nego_token)
        result = parse_ts_request(original)
        assert "nego_token" in result
        # The nego token should be recoverable


class TestCredSSPAuth:
    """Tests for CredSSPAuth class."""

    def test_credssp_auth_initialization(self) -> None:
        """Test CredSSPAuth can be initialized."""
        from simple_rdp.credssp import CredSSPAuth

        auth = CredSSPAuth(
            hostname="testserver",
            username="testuser",
            password="testpass",
            domain="DOMAIN",
        )
        assert auth.client_nonce is not None
        assert len(auth.client_nonce) == NONCE_SIZE
        assert auth.server_version == CREDSSP_VERSION

    def test_credssp_auth_client_nonce(self) -> None:
        """Test CredSSPAuth generates random client nonce."""
        from simple_rdp.credssp import CredSSPAuth

        auth1 = CredSSPAuth(hostname="test", username="u", password="p")
        auth2 = CredSSPAuth(hostname="test", username="u", password="p")
        # Nonces should be different (random)
        assert auth1.client_nonce != auth2.client_nonce

    def test_credssp_auth_server_version_setter(self) -> None:
        """Test CredSSPAuth server version setter."""
        from simple_rdp.credssp import CredSSPAuth

        auth = CredSSPAuth(hostname="test", username="u", password="p")
        auth.server_version = 3
        assert auth.server_version == 3

    def test_credssp_auth_pending_token_initial(self) -> None:
        """Test CredSSPAuth pending token is None initially."""
        from simple_rdp.credssp import CredSSPAuth

        auth = CredSSPAuth(hostname="test", username="u", password="p")
        assert auth.pending_token is None

    def test_credssp_auth_get_initial_token(self) -> None:
        """Test CredSSPAuth can get initial token."""
        from simple_rdp.credssp import CredSSPAuth

        auth = CredSSPAuth(
            hostname="testserver",
            username="testuser",
            password="testpass",
        )
        token = auth.get_initial_token()
        assert isinstance(token, bytes)
        assert len(token) > 0


class TestTsRequestWithNonce:
    """Tests for TSRequest with client nonce."""

    def test_build_ts_request_with_nonce(self) -> None:
        """Test building TSRequest with client nonce."""
        pub_key_auth = b"encrypted_pub_key"
        client_nonce = b"a" * NONCE_SIZE
        result = build_ts_request_with_pub_key_auth(
            pub_key_auth=pub_key_auth,
            client_nonce=client_nonce,
        )
        assert isinstance(result, bytes)
        assert len(result) > len(pub_key_auth) + NONCE_SIZE

    def test_build_ts_request_with_nego_and_nonce(self) -> None:
        """Test building TSRequest with nego token and nonce."""
        pub_key_auth = b"encrypted_pub_key"
        nego_token = b"nego_token_data"
        client_nonce = b"b" * NONCE_SIZE
        result = build_ts_request_with_pub_key_auth(
            pub_key_auth=pub_key_auth,
            nego_token=nego_token,
            client_nonce=client_nonce,
        )
        assert isinstance(result, bytes)
        assert len(result) > len(pub_key_auth) + len(nego_token) + NONCE_SIZE


class TestAsn1EncodingEdgeCases:
    """Tests for ASN.1 encoding edge cases."""

    def test_encode_asn1_length_boundary_127(self) -> None:
        """Test encoding length at boundary of 127."""
        assert _encode_asn1_length(127) == bytes([127])

    def test_encode_asn1_length_boundary_128(self) -> None:
        """Test encoding length at boundary of 128."""
        assert _encode_asn1_length(128) == bytes([0x81, 128])

    def test_encode_asn1_length_boundary_255(self) -> None:
        """Test encoding length at boundary of 255."""
        assert _encode_asn1_length(255) == bytes([0x81, 255])

    def test_encode_asn1_length_boundary_256(self) -> None:
        """Test encoding length at boundary of 256."""
        assert _encode_asn1_length(256) == bytes([0x82, 0x01, 0x00])

    def test_encode_asn1_context_higher_tag(self) -> None:
        """Test encoding context with higher tag numbers."""
        from simple_rdp.credssp import ASN1_CONTEXT_1
        from simple_rdp.credssp import ASN1_CONTEXT_2

        content = b"\x00\x01\x02"
        result1 = _encode_asn1_context(ASN1_CONTEXT_1, content)
        assert result1[0] == ASN1_CONTEXT_1

        result2 = _encode_asn1_context(ASN1_CONTEXT_2, content)
        assert result2[0] == ASN1_CONTEXT_2


class TestParseTsRequestAdvanced:
    """Advanced tests for TSRequest parsing to cover all branches."""

    def test_parse_ts_request_with_pub_key_auth(self) -> None:
        """Test parsing TSRequest with pubKeyAuth field."""
        # Build a request with pubKeyAuth and parse it back
        pub_key_auth = b"test_pub_key_auth_data"
        original = build_ts_request_with_pub_key_auth(pub_key_auth=pub_key_auth)
        result = parse_ts_request(original)
        assert "pub_key_auth" in result
        assert result["pub_key_auth"] == pub_key_auth

    def test_parse_ts_request_with_client_nonce(self) -> None:
        """Test parsing TSRequest with clientNonce field."""
        pub_key_auth = b"encrypted_key"
        client_nonce = b"c" * NONCE_SIZE
        original = build_ts_request_with_pub_key_auth(
            pub_key_auth=pub_key_auth,
            client_nonce=client_nonce,
        )
        result = parse_ts_request(original)
        assert "client_nonce" in result
        assert result["client_nonce"] == client_nonce

    def test_parse_ts_request_with_auth_info(self) -> None:
        """Test parsing TSRequest with authInfo field."""
        auth_info = b"encrypted_creds_data"
        original = build_ts_request_with_credentials(auth_info=auth_info)
        result = parse_ts_request(original)
        assert "auth_info" in result
        assert result["auth_info"] == auth_info

    def test_parse_ts_request_invalid_not_sequence(self) -> None:
        """Test parsing TSRequest fails if not a SEQUENCE."""
        # Build invalid data that's not a SEQUENCE
        invalid_data = bytes([ASN1_INTEGER, 0x01, 0x05])
        with pytest.raises(ValueError, match="Expected SEQUENCE"):
            parse_ts_request(invalid_data)


class TestCredSSPAuthAdvanced:
    """Advanced tests for CredSSPAuth class."""

    def test_credssp_auth_complete_attribute(self) -> None:
        """Test CredSSPAuth complete attribute."""
        from simple_rdp.credssp import CredSSPAuth

        auth = CredSSPAuth(hostname="test", username="u", password="p")
        # Initially, complete is False (depends on underlying SPNEGO context)
        # This just accesses the property
        _ = auth.complete

    def test_credssp_auth_hostname_stored(self) -> None:
        """Test CredSSPAuth stores hostname."""
        from simple_rdp.credssp import CredSSPAuth

        auth = CredSSPAuth(hostname="myserver.local", username="u", password="p")
        assert auth._hostname == "myserver.local"

    def test_credssp_auth_domain_stored(self) -> None:
        """Test CredSSPAuth stores domain."""
        from simple_rdp.credssp import CredSSPAuth

        auth = CredSSPAuth(hostname="test", username="u", password="p", domain="MYDOMAIN")
        assert auth._domain == "MYDOMAIN"


class TestAsn1LengthLarge:
    """Tests for ASN.1 length encoding with large values."""

    def test_encode_asn1_length_large_3byte(self) -> None:
        """Test encoding length requiring 3-byte form."""
        # 0x10000 = 65536
        result = _encode_asn1_length(0x10000)
        # Should use 0x83 prefix for 3-byte length
        assert result[0] == 0x83
        assert len(result) == 4

    def test_encode_asn1_length_medium_large(self) -> None:
        """Test encoding length near 2-byte boundary."""
        result = _encode_asn1_length(0xFFFF)  # 65535
        assert result[0] == 0x82
        assert result[1] == 0xFF
        assert result[2] == 0xFF


class TestCredSSPAuthComputeHash:
    """Tests for CredSSPAuth hash computation methods."""

    def test_compute_client_server_hash(self) -> None:
        """Test computing client-to-server hash."""
        from hashlib import sha256

        from simple_rdp.credssp import CLIENT_SERVER_HASH_MAGIC
        from simple_rdp.credssp import CredSSPAuth

        auth = CredSSPAuth(hostname="test", username="u", password="p")
        public_key = b"test_public_key_data"

        # Compute expected hash
        expected = sha256(CLIENT_SERVER_HASH_MAGIC + auth.client_nonce + public_key).digest()
        result = auth.compute_client_server_hash(public_key)

        assert result == expected

    def test_compute_server_client_hash(self) -> None:
        """Test computing server-to-client hash."""
        from hashlib import sha256

        from simple_rdp.credssp import SERVER_CLIENT_HASH_MAGIC
        from simple_rdp.credssp import CredSSPAuth

        auth = CredSSPAuth(hostname="test", username="u", password="p")
        public_key = b"test_public_key_data"

        # Compute expected hash
        expected = sha256(SERVER_CLIENT_HASH_MAGIC + auth.client_nonce + public_key).digest()
        result = auth.compute_server_client_hash(public_key)

        assert result == expected


class TestCredSSPAuthSetters:
    """Tests for CredSSPAuth property setters."""

    def test_set_server_public_key(self) -> None:
        """Test setting server public key."""
        from simple_rdp.credssp import CredSSPAuth

        auth = CredSSPAuth(hostname="test", username="u", password="p")
        public_key = b"server_public_key"
        auth.set_server_public_key(public_key)
        assert auth._server_public_key == public_key


class TestTsCredentialsBuild:
    """Tests for building TS credentials."""

    def test_build_ts_credentials_unicode(self) -> None:
        """Test building TS credentials with unicode strings."""
        result = build_ts_credentials(domain="MYDOMAIN", username="myuser", password="mypassword123!")
        assert isinstance(result, bytes)
        # Should contain the encoded strings
        assert len(result) > 0

    def test_build_ts_credentials_empty_domain(self) -> None:
        """Test building TS credentials with empty domain."""
        result = build_ts_credentials(domain="", username="user", password="pass")
        assert isinstance(result, bytes)


class TestAsn1IntegerEdgeCases:
    """Edge case tests for ASN.1 integer encoding."""

    def test_encode_integer_exactly_128(self) -> None:
        """Test encoding integer 128 (boundary case)."""
        result = _encode_asn1_integer(128)
        assert result[0] == ASN1_INTEGER
        # 128 needs padding to avoid negative interpretation
        assert len(result) >= 3

    def test_encode_integer_255(self) -> None:
        """Test encoding integer 255."""
        result = _encode_asn1_integer(255)
        assert result[0] == ASN1_INTEGER
        assert len(result) >= 3

    def test_encode_integer_256(self) -> None:
        """Test encoding integer 256 (2 bytes needed)."""
        result = _encode_asn1_integer(256)
        assert result[0] == ASN1_INTEGER
        assert len(result) >= 4



"""
Pointer visibility test - Verifies cursor compositing in recordings.

This test specifically tests pointer visibility with clear movement patterns:
- Rectangular path around screen edges
- Circular motion in center
- Diagonal selection (drag) patterns

The pointer should be visible in the recording at all times.
"""

import asyncio
import logging
import math
import os
from datetime import datetime
from pathlib import Path

from dotenv import load_dotenv
from rich.logging import RichHandler

from simple_rdp import RDPClient

load_dotenv()


async def rectangular_pattern(client: RDPClient, width: int, height: int) -> None:
    """
    Move mouse in a rectangular pattern around screen edges.

    This creates a clear, visible path that's easy to verify in recordings.
    """
    print("\n  === Rectangular Pattern ===")
    margin = 100
    steps = 20
    delay = 0.1

    corners = [
        (margin, margin),  # Top-left
        (width - margin, margin),  # Top-right
        (width - margin, height - margin),  # Bottom-right
        (margin, height - margin),  # Bottom-left
        (margin, margin),  # Back to top-left
    ]

    for i in range(len(corners) - 1):
        x1, y1 = corners[i]
        x2, y2 = corners[i + 1]

        for step in range(steps + 1):
            t = step / steps
            x = int(x1 + (x2 - x1) * t)
            y = int(y1 + (y2 - y1) * t)
            await client.mouse_move(x, y)
            await asyncio.sleep(delay)

        print(f"    Corner {i + 1} -> {i + 2}: ({x2}, {y2})")


async def circular_pattern(client: RDPClient, width: int, height: int) -> None:
    """
    Move mouse in a circular pattern in the center of the screen.

    Creates a smooth circular motion that's visually distinct.
    """
    print("\n  === Circular Pattern ===")
    cx, cy = width // 2, height // 2
    radius = 200
    steps = 60  # Full circle
    delay = 0.05

    for step in range(steps + 1):
        angle = (step / steps) * 2 * math.pi
        x = int(cx + radius * math.cos(angle))
        y = int(cy + radius * math.sin(angle))
        await client.mouse_move(x, y)
        await asyncio.sleep(delay)

    print(f"    Completed circle at center ({cx}, {cy}), radius {radius}")


async def diagonal_selection_pattern(client: RDPClient, width: int, height: int) -> None:
    """
    Perform diagonal selection (drag) patterns.

    Simulates selection boxes being drawn diagonally across the screen.
    """
    print("\n  === Diagonal Selection Pattern ===")
    steps = 30
    delay = 0.05

    # Diagonal from top-left to center (drag)
    print("    Selection 1: Top-left to center")
    await client.mouse_move(100, 100)
    await asyncio.sleep(0.2)
    await client.mouse_button_down(100, 100, button="left")
    await asyncio.sleep(0.1)

    for step in range(steps + 1):
        t = step / steps
        x = int(100 + (width // 2 - 100) * t)
        y = int(100 + (height // 2 - 100) * t)
        await client.mouse_move(x, y)
        await asyncio.sleep(delay)

    await client.mouse_button_up(width // 2, height // 2, button="left")
    await asyncio.sleep(0.3)

    # Diagonal from top-right to bottom-left (drag)
    print("    Selection 2: Top-right to bottom-left")
    await client.mouse_move(width - 100, 100)
    await asyncio.sleep(0.2)
    await client.mouse_button_down(width - 100, 100, button="left")
    await asyncio.sleep(0.1)

    for step in range(steps + 1):
        t = step / steps
        x = int((width - 100) - (width - 200) * t)
        y = int(100 + (height - 200) * t)
        await client.mouse_move(x, y)
        await asyncio.sleep(delay)

    await client.mouse_button_up(100, height - 100, button="left")
    await asyncio.sleep(0.3)


async def spiral_pattern(client: RDPClient, width: int, height: int) -> None:
    """
    Move mouse in a spiral pattern from center outward.

    Creates an expanding spiral that's visually interesting.
    """
    print("\n  === Spiral Pattern ===")
    cx, cy = width // 2, height // 2
    max_radius = min(width, height) // 3
    rotations = 3
    points_per_rotation = 40
    total_points = rotations * points_per_rotation
    delay = 0.03

    for point in range(total_points + 1):
        t = point / total_points
        radius = max_radius * t
        angle = t * rotations * 2 * math.pi
        x = int(cx + radius * math.cos(angle))
        y = int(cy + radius * math.sin(angle))
        await client.mouse_move(x, y)
        await asyncio.sleep(delay)

    print(f"    Completed spiral with {rotations} rotations")


async def perform_pointer_test(client: RDPClient) -> None:
    """
    Perform all pointer visibility tests.
    """
    print("\n=== Pointer Visibility Test ===")
    print("Testing pointer compositing with various movement patterns...")

    # Screen dimensions (assuming 1920x1080)
    width, height = 1920, 1080

    # Initial position - center of screen
    print("\n  Moving pointer to center...")
    await client.mouse_move(width // 2, height // 2)
    await asyncio.sleep(1)

    # Pattern 1: Rectangle
    await rectangular_pattern(client, width, height)
    await asyncio.sleep(0.5)

    # Pattern 2: Circle
    await circular_pattern(client, width, height)
    await asyncio.sleep(0.5)

    # Pattern 3: Spiral
    await spiral_pattern(client, width, height)
    await asyncio.sleep(0.5)

    # Pattern 4: Diagonal selections
    await diagonal_selection_pattern(client, width, height)
    await asyncio.sleep(0.5)

    # Final position - center
    print("\n  Returning pointer to center...")
    await client.mouse_move(width // 2, height // 2)
    await asyncio.sleep(1)

    print("\n=== Pointer Test Complete ===\n")


async def main() -> None:
    logging.basicConfig(level=logging.INFO, format="%(message)s", handlers=[RichHandler()])

    # Get RDP connection details from environment
    host = os.environ.get("RDP_HOST", "")
    username = os.environ.get("RDP_USER", "")
    password = os.environ.get("RDP_PASS", "")

    if not host or not username or not password:
        print("ERROR: Set RDP_HOST, RDP_USER, RDP_PASS environment variables (or use .env file)")
        return

    # Create session directory with timestamp
    timestamp = datetime.now().strftime("%Y%m%d_%H%M%S")
    session_dir = Path("tests/e2e/sessions") / f"pointer_test_{timestamp}"
    session_dir.mkdir(parents=True, exist_ok=True)
    video_ts_path = session_dir / "pointer_test.ts"
    video_mp4_path = session_dir / "pointer_test.mp4"

    print(f"\n{'=' * 60}")
    print(f"Pointer Visibility Test - {timestamp}")
    print(f"Session directory: {session_dir}")
    print(f"Video output: {video_ts_path}")
    print(f"{'=' * 60}\n")

    async with RDPClient(host=host, username=username, password=password, show_wallpaper=True) as client:
        print("Connected to RDP server!\n")

        # Wait for initial screen to render
        await asyncio.sleep(2)

        # Start recording
        print("Starting recording...")
        await client.start_file_recording(str(video_ts_path))
        print(f"Recording to: {video_ts_path}")

        try:
            # Perform pointer visibility tests
            await perform_pointer_test(client)

        finally:
            # Stop recording
            print("Stopping recording...")
            await client.stop_file_recording()
            await client.stop_streaming()

    # Show results
    print(f"\n{'=' * 60}")
    print("Recording complete!")
    print(f"{'=' * 60}\n")

    if video_ts_path.exists():
        size_mb = video_ts_path.stat().st_size / (1024 * 1024)
        stats = client.get_recording_stats()
        print(f"✓ Video saved: {video_ts_path}")
        print(f"  Size: {size_mb:.2f} MB")
        print(f"  Frames encoded: {stats.get('frames_encoded', 'N/A')}")
        print(f"  Pointer updates: {stats.get('pointer_updates', 'N/A')}")
        print(f"  Pointer throttled: {stats.get('pointer_updates_throttled', 'N/A')}")

        # Transcode to MP4
        print("\nTranscoding to MP4...")
        if RDPClient.transcode(str(video_ts_path), str(video_mp4_path)):
            mp4_size_mb = video_mp4_path.stat().st_size / (1024 * 1024)
            print(f"✓ MP4 saved: {video_mp4_path}")
            print(f"  Size: {mp4_size_mb:.2f} MB")
        else:
            print("✗ Transcode failed")
    else:
        print("✗ Video file not found")

    print(f"\n{'=' * 60}")
    print("To verify pointer visibility:")
    print(f"  1. Open {video_mp4_path}")
    print("  2. Look for the white arrow cursor during movement patterns")
    print("  3. Cursor should be visible in rectangle, circle, and diagonal patterns")
    print(f"{'=' * 60}\n")


if __name__ == "__main__":
    asyncio.run(main())



"""Tests for MCS layer module."""

import struct

import pytest

from simple_rdp.mcs import CS_CLUSTER
from simple_rdp.mcs import CS_CORE
from simple_rdp.mcs import CS_NET
from simple_rdp.mcs import CS_SECURITY
from simple_rdp.mcs import GCC_OBJECT_ID
from simple_rdp.mcs import H221_CS_KEY
from simple_rdp.mcs import MCS_TYPE_ATTACH_USER_REQUEST
from simple_rdp.mcs import MCS_TYPE_CHANNEL_JOIN_REQUEST
from simple_rdp.mcs import MCS_TYPE_CONNECT_INITIAL
from simple_rdp.mcs import MCS_TYPE_ERECT_DOMAIN_REQUEST
from simple_rdp.mcs import MCS_TYPE_SEND_DATA_REQUEST
from simple_rdp.mcs import _ber_write_application_tag
from simple_rdp.mcs import _ber_write_boolean
from simple_rdp.mcs import _ber_write_integer
from simple_rdp.mcs import _ber_write_length
from simple_rdp.mcs import _ber_write_octet_string
from simple_rdp.mcs import _ber_write_sequence
from simple_rdp.mcs import _per_write_integer
from simple_rdp.mcs import _per_write_length
from simple_rdp.mcs import build_client_cluster_data
from simple_rdp.mcs import build_client_core_data
from simple_rdp.mcs import build_client_network_data
from simple_rdp.mcs import build_client_security_data
from simple_rdp.mcs import build_domain_parameters
from simple_rdp.mcs import build_gcc_conference_create_request
from simple_rdp.mcs import build_gcc_connect_data
from simple_rdp.mcs import build_mcs_attach_user_request
from simple_rdp.mcs import build_mcs_channel_join_request
from simple_rdp.mcs import build_mcs_connect_initial
from simple_rdp.mcs import build_mcs_erect_domain_request
from simple_rdp.mcs import build_mcs_send_data_request
from simple_rdp.mcs import parse_mcs_attach_user_confirm
from simple_rdp.mcs import parse_mcs_channel_join_confirm
from simple_rdp.mcs import parse_mcs_connect_response


class TestBerEncoding:
    """Tests for BER encoding functions."""

    def test_ber_write_length_short_form(self) -> None:
        """Test BER length encoding - short form."""
        assert _ber_write_length(0) == bytes([0])
        assert _ber_write_length(127) == bytes([127])
        assert _ber_write_length(50) == bytes([50])

    def test_ber_write_length_long_form_1byte(self) -> None:
        """Test BER length encoding - long form 1 byte."""
        assert _ber_write_length(128) == bytes([0x81, 128])
        assert _ber_write_length(255) == bytes([0x81, 255])

    def test_ber_write_length_long_form_2bytes(self) -> None:
        """Test BER length encoding - long form 2 bytes."""
        result = _ber_write_length(256)
        assert result == bytes([0x82, 0x01, 0x00])
        result = _ber_write_length(0xFFFF)
        assert result == bytes([0x82, 0xFF, 0xFF])

    def test_ber_write_length_too_large(self) -> None:
        """Test BER length encoding - too large."""
        with pytest.raises(ValueError, match="Length too large"):
            _ber_write_length(0x10000)

    def test_ber_write_integer_zero(self) -> None:
        """Test BER integer encoding - zero."""
        result = _ber_write_integer(0)
        assert result == bytes([0x02, 0x01, 0x00])

    def test_ber_write_integer_small(self) -> None:
        """Test BER integer encoding - small value."""
        result = _ber_write_integer(100)
        assert result[0] == 0x02  # INTEGER tag
        assert result[1] == 1  # Length
        assert result[2] == 100

    def test_ber_write_integer_large(self) -> None:
        """Test BER integer encoding - large value."""
        result = _ber_write_integer(65535)
        assert result[0] == 0x02  # INTEGER tag
        # Value needs 3 bytes (leading 0 to avoid negative interpretation)

    def test_ber_write_integer_negative_raises(self) -> None:
        """Test BER integer encoding - negative raises."""
        with pytest.raises(ValueError, match="Negative integers"):
            _ber_write_integer(-1)

    def test_ber_write_integer_high_bit_set(self) -> None:
        """Test BER integer encoding - value with high bit set needs padding."""
        # 0x80 = 128, has high bit set, needs leading 0x00
        result = _ber_write_integer(128)
        assert result[0] == 0x02  # INTEGER tag
        # Should have 2 bytes of content: 0x00 0x80
        assert result[1] == 2  # Length
        assert result[2] == 0x00  # Leading zero
        assert result[3] == 0x80

    def test_ber_write_integer_0xFF(self) -> None:
        """Test BER integer encoding - 0xFF needs padding."""
        result = _ber_write_integer(0xFF)
        assert result[0] == 0x02  # INTEGER tag
        assert result[2] == 0x00  # Leading zero for high bit

    def test_ber_write_octet_string(self) -> None:
        """Test BER octet string encoding."""
        data = b"hello"
        result = _ber_write_octet_string(data)
        assert result[0] == 0x04  # OCTET STRING tag
        assert result[1] == len(data)  # Length
        assert result[2:] == data

    def test_ber_write_boolean_true(self) -> None:
        """Test BER boolean encoding - true."""
        result = _ber_write_boolean(True)
        assert result == bytes([0x01, 0x01, 0xFF])

    def test_ber_write_boolean_false(self) -> None:
        """Test BER boolean encoding - false."""
        result = _ber_write_boolean(False)
        assert result == bytes([0x01, 0x01, 0x00])

    def test_ber_write_sequence(self) -> None:
        """Test BER sequence encoding."""
        content = b"\x01\x02\x03"
        result = _ber_write_sequence(content)
        assert result[0] == 0x30  # SEQUENCE tag
        assert result[1] == len(content)
        assert result[2:] == content


class TestPerEncoding:
    """Tests for PER encoding functions."""

    def test_per_write_length_short(self) -> None:
        """Test PER length encoding - short."""
        assert _per_write_length(0) == bytes([0])
        assert _per_write_length(127) == bytes([127])

    def test_per_write_length_long(self) -> None:
        """Test PER length encoding - long."""
        result = _per_write_length(128)
        assert result == bytes([0x80, 128])

    def test_per_write_length_too_large(self) -> None:
        """Test PER length encoding - too large."""
        with pytest.raises(ValueError, match="Length too large for PER"):
            _per_write_length(0x4000)


class TestDomainParameters:
    """Tests for domain parameters building."""

    def test_build_domain_parameters_defaults(self) -> None:
        """Test building domain parameters with defaults."""
        result = build_domain_parameters()
        assert len(result) > 0
        assert isinstance(result, bytes)

    def test_build_domain_parameters_custom(self) -> None:
        """Test building domain parameters with custom values."""
        result = build_domain_parameters(
            max_channel_ids=100,
            max_user_ids=10,
            max_mcs_pdu_size=32768,
        )
        assert len(result) > 0


class TestMcsPduBuilding:
    """Tests for MCS PDU building functions."""

    def test_build_erect_domain_request(self) -> None:
        """Test building erect domain request."""
        result = build_mcs_erect_domain_request()
        assert len(result) > 0
        assert result[0] == MCS_TYPE_ERECT_DOMAIN_REQUEST

    def test_build_attach_user_request(self) -> None:
        """Test building attach user request."""
        result = build_mcs_attach_user_request()
        assert len(result) > 0
        assert result[0] == MCS_TYPE_ATTACH_USER_REQUEST

    def test_build_channel_join_request(self) -> None:
        """Test building channel join request."""
        user_id = 1001
        channel_id = 1003
        result = build_mcs_channel_join_request(user_id, channel_id)
        assert len(result) > 0
        assert result[0] == MCS_TYPE_CHANNEL_JOIN_REQUEST


class TestConstants:
    """Tests for MCS constants."""

    def test_gcc_object_id(self) -> None:
        """Test GCC object ID format."""
        assert len(GCC_OBJECT_ID) == 5
        assert isinstance(GCC_OBJECT_ID, bytes)

    def test_h221_cs_key(self) -> None:
        """Test H.221 client-to-server key."""
        assert H221_CS_KEY == b"Duca"

    def test_user_data_types(self) -> None:
        """Test user data type constants."""
        assert CS_CORE == 0xC001
        assert CS_SECURITY == 0xC002
        assert CS_NET == 0xC003

    def test_cluster_type(self) -> None:
        """Test cluster type constant."""
        assert CS_CLUSTER == 0xC004

    def test_mcs_type_constants(self) -> None:
        """Test MCS type constants."""
        assert MCS_TYPE_CONNECT_INITIAL == 0x65
        assert MCS_TYPE_ERECT_DOMAIN_REQUEST == 0x04
        assert MCS_TYPE_ATTACH_USER_REQUEST == 0x28
        assert MCS_TYPE_CHANNEL_JOIN_REQUEST == 0x38
        assert MCS_TYPE_SEND_DATA_REQUEST == 0x64

    def test_mcs_confirm_type_constants(self) -> None:
        """Test MCS confirm type constants."""
        from simple_rdp.mcs import MCS_TYPE_ATTACH_USER_CONFIRM
        from simple_rdp.mcs import MCS_TYPE_CHANNEL_JOIN_CONFIRM
        from simple_rdp.mcs import MCS_TYPE_CONNECT_RESPONSE

        assert MCS_TYPE_CONNECT_RESPONSE == 0x66
        assert MCS_TYPE_ATTACH_USER_CONFIRM == 0x2E
        assert MCS_TYPE_CHANNEL_JOIN_CONFIRM == 0x3E

    def test_server_user_data_types(self) -> None:
        """Test server user data type constants."""
        from simple_rdp.mcs import SC_CORE
        from simple_rdp.mcs import SC_NET
        from simple_rdp.mcs import SC_SECURITY

        assert SC_CORE == 0x0C01
        assert SC_SECURITY == 0x0C02
        assert SC_NET == 0x0C03


class TestBerApplicationTag:
    """Tests for BER application tag encoding."""

    def test_ber_write_application_tag_small(self) -> None:
        """Test application tag with small tag number."""
        content = b"\x01\x02"
        result = _ber_write_application_tag(5, content)
        assert result[0] == 0x65  # 0x60 | 5
        assert len(result) > 2

    def test_ber_write_application_tag_large(self) -> None:
        """Test application tag with large tag number (>30)."""
        content = b"\x01\x02"
        result = _ber_write_application_tag(101, content)  # MCS_TYPE_CONNECT_INITIAL
        assert result[0] == 0x7F  # Multi-byte encoding


class TestPerIntegerEncoding:
    """Tests for PER integer encoding."""

    def test_per_write_integer_small(self) -> None:
        """Test PER integer encoding - small value."""
        result = _per_write_integer(0)
        assert isinstance(result, bytes)

    def test_per_write_integer_medium(self) -> None:
        """Test PER integer encoding - medium value."""
        result = _per_write_integer(256)
        assert isinstance(result, bytes)
        assert len(result) >= 2

    def test_per_write_integer_large_2byte(self) -> None:
        """Test PER integer encoding - value requiring 2 bytes."""
        result = _per_write_integer(1000)
        assert isinstance(result, bytes)
        # Should be 2-byte format: 0x02, high, low
        assert result[0] == 0x02
        assert len(result) == 3

    def test_per_write_integer_max_2byte(self) -> None:
        """Test PER integer encoding - max 2-byte value."""
        result = _per_write_integer(65535)
        assert isinstance(result, bytes)
        assert result[0] == 0x02
        assert result[1] == 0xFF
        assert result[2] == 0xFF

    def test_per_write_integer_too_large(self) -> None:
        """Test PER integer encoding - too large raises error."""
        with pytest.raises(ValueError, match="Integer too large"):
            _per_write_integer(65536)


class TestClientCoreData:
    """Tests for Client Core Data building."""

    def test_build_client_core_data_defaults(self) -> None:
        """Test building client core data with defaults."""
        result = build_client_core_data()
        assert isinstance(result, bytes)
        assert len(result) > 100
        # Check header type
        header_type = struct.unpack("<H", result[:2])[0]
        assert header_type == CS_CORE

    def test_build_client_core_data_custom_resolution(self) -> None:
        """Test building client core data with custom resolution."""
        result = build_client_core_data(desktop_width=1280, desktop_height=720)
        assert isinstance(result, bytes)
        # Width and height are after header (4 bytes) and version (4 bytes)
        width = struct.unpack("<H", result[8:10])[0]
        height = struct.unpack("<H", result[10:12])[0]
        assert width == 1280
        assert height == 720

    def test_build_client_core_data_custom_client_name(self) -> None:
        """Test building client core data with custom client name."""
        result = build_client_core_data(client_name="testclient")
        assert isinstance(result, bytes)


class TestClientSecurityData:
    """Tests for Client Security Data building."""

    def test_build_client_security_data_defaults(self) -> None:
        """Test building client security data with defaults."""
        result = build_client_security_data()
        assert isinstance(result, bytes)
        # Header type
        header_type = struct.unpack("<H", result[:2])[0]
        assert header_type == CS_SECURITY
        # Fixed size of 12 bytes
        length = struct.unpack("<H", result[2:4])[0]
        assert length == 12

    def test_build_client_security_data_custom(self) -> None:
        """Test building client security data with custom encryption."""
        result = build_client_security_data(encryption_methods=0x0B)
        assert isinstance(result, bytes)
        encryption = struct.unpack("<I", result[4:8])[0]
        assert encryption == 0x0B


class TestClientNetworkData:
    """Tests for Client Network Data building."""

    def test_build_client_network_data_no_channels(self) -> None:
        """Test building client network data without channels."""
        result = build_client_network_data()
        assert isinstance(result, bytes)
        header_type = struct.unpack("<H", result[:2])[0]
        assert header_type == CS_NET

    def test_build_client_network_data_with_channels(self) -> None:
        """Test building client network data with channels."""
        channels = [("rdpdr", 0x80000000), ("cliprdr", 0xC0000000)]
        result = build_client_network_data(channels=channels)
        assert isinstance(result, bytes)
        # Should include channel definitions
        assert len(result) > 8


class TestClientClusterData:
    """Tests for Client Cluster Data building."""

    def test_build_client_cluster_data_defaults(self) -> None:
        """Test building client cluster data with defaults."""
        result = build_client_cluster_data()
        assert isinstance(result, bytes)
        header_type = struct.unpack("<H", result[:2])[0]
        assert header_type == CS_CLUSTER


class TestGccConference:
    """Tests for GCC conference building."""

    def test_build_gcc_conference_create_request(self) -> None:
        """Test building GCC conference create request."""
        user_data = build_client_core_data()
        result = build_gcc_conference_create_request(user_data)
        assert isinstance(result, bytes)
        assert len(result) > len(user_data)

    def test_build_gcc_connect_data(self) -> None:
        """Test building GCC connect data."""
        gcc_ccr = build_gcc_conference_create_request(build_client_core_data())
        result = build_gcc_connect_data(gcc_ccr)
        assert isinstance(result, bytes)
        assert len(result) > len(gcc_ccr)


class TestMcsConnectInitial:
    """Tests for MCS Connect Initial building."""

    def test_build_mcs_connect_initial_defaults(self) -> None:
        """Test building MCS connect initial with defaults."""
        # Build user data (client core + security + network)
        user_data = build_client_core_data() + build_client_security_data() + build_client_network_data()
        result = build_mcs_connect_initial(user_data)
        assert isinstance(result, bytes)
        assert len(result) > 100
        # Check APPLICATION tag (0x7F for multi-byte, then 0x65 for 101)
        assert result[0] == 0x7F

    def test_build_mcs_connect_initial_custom_resolution(self) -> None:
        """Test building MCS connect initial with custom resolution."""
        user_data = build_client_core_data(desktop_width=800, desktop_height=600)
        result = build_mcs_connect_initial(user_data)
        assert isinstance(result, bytes)


class TestMcsSendDataRequest:
    """Tests for MCS Send Data Request building."""

    def test_build_mcs_send_data_request(self) -> None:
        """Test building MCS send data request."""
        user_data = b"\x00\x01\x02\x03"
        result = build_mcs_send_data_request(
            user_id=1001,
            channel_id=1003,
            user_data=user_data,
        )
        assert isinstance(result, bytes)
        # Type byte has data priority in lower 2 bits
        assert (result[0] & 0xFC) == MCS_TYPE_SEND_DATA_REQUEST


class TestMcsParsingFunctions:
    """Tests for MCS parsing functions."""

    def test_parse_mcs_attach_user_confirm_valid(self) -> None:
        """Test parsing valid attach user confirm."""
        # Build a valid attach user confirm
        # Byte 0: bits 7-2 = choice (11 for attachUserConfirm), bit 1 = initiator present (1)
        # So (11 << 2) | 0x02 = 0x2E
        # Byte 1: result value (0 = success)
        # Bytes 2-3: user_id - 1001 (big-endian)
        data = bytes([0x2E, 0x00, 0x00, 0x00])  # Type, result, user_id offset = 0
        result = parse_mcs_attach_user_confirm(data)
        assert "result" in result
        assert result["result"] == 0

    def test_parse_mcs_attach_user_confirm_with_user_id(self) -> None:
        """Test parsing attach user confirm with user ID."""
        # User ID present, user_id = 1005 => offset = 4
        data = bytes([0x2E, 0x00, 0x00, 0x04])
        result = parse_mcs_attach_user_confirm(data)
        assert result["user_id"] == 1005

    def test_parse_mcs_attach_user_confirm_too_short(self) -> None:
        """Test parsing attach user confirm with too short data."""
        with pytest.raises(ValueError, match="too short"):
            parse_mcs_attach_user_confirm(b"\x2e")

    def test_parse_mcs_channel_join_confirm_valid(self) -> None:
        """Test parsing valid channel join confirm."""
        # Type: 0x3E (CHANNEL_JOIN_CONFIRM), result: 0, initiator: 1001, channel: 1003
        # Byte 0: (15 << 2) | 0x02 = 0x3E (type + channel present)
        data = bytes(
            [
                0x3E,  # Type + channel present
                0x00,  # Result (success)
                0x00,
                0x00,  # Initiator offset (0 => 1001)
                0x03,
                0xEB,  # Requested channel (1003)
                0x03,
                0xEB,  # Joined channel (1003)
            ]
        )
        result = parse_mcs_channel_join_confirm(data)
        assert isinstance(result, dict)
        assert result["result"] == 0
        assert result["user_id"] == 1001
        assert result["channel_id"] == 1003

    def test_parse_mcs_channel_join_confirm_too_short(self) -> None:
        """Test parsing channel join confirm with too short data."""
        with pytest.raises(ValueError, match="too short"):
            parse_mcs_channel_join_confirm(bytes([0x3E, 0x00, 0x00, 0x00, 0x03]))

    def test_parse_mcs_channel_join_confirm_wrong_type(self) -> None:
        """Test parsing channel join confirm with wrong type."""
        # Type byte doesn't match expected
        data = bytes([0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00])
        with pytest.raises(ValueError, match="Expected Channel Join Confirm"):
            parse_mcs_channel_join_confirm(data)


class TestMcsConnectResponseParsing:
    """Tests for MCS Connect Response parsing."""

    def test_parse_mcs_connect_response_wrong_tag(self) -> None:
        """Test parsing MCS connect response with wrong APPLICATION tag."""
        # Wrong tag - not 0x7F 0x66
        data = bytes([0x00, 0x00, 0x00])
        with pytest.raises(ValueError, match="Expected MCS Connect Response"):
            parse_mcs_connect_response(data)

    def test_parse_mcs_connect_response_wrong_type(self) -> None:
        """Test parsing MCS connect response with wrong type byte."""
        # Right 0x7F but wrong type (0x65 instead of 0x66)
        data = bytes([0x7F, 0x65, 0x00])
        with pytest.raises(ValueError, match="Expected MCS Connect Response"):
            parse_mcs_connect_response(data)


class TestMcsSendDataRequestLengths:
    """Tests for MCS Send Data Request with different lengths."""

    def test_build_mcs_send_data_request_short_data(self) -> None:
        """Test building MCS send data request with short data."""
        user_data = b"\x00" * 50
        result = build_mcs_send_data_request(
            user_id=1001,
            channel_id=1003,
            user_data=user_data,
        )
        assert isinstance(result, bytes)
        assert len(result) > len(user_data)

    def test_build_mcs_send_data_request_medium_data(self) -> None:
        """Test building MCS send data request with medium-length data."""
        user_data = b"\x00" * 200
        result = build_mcs_send_data_request(
            user_id=1001,
            channel_id=1003,
            user_data=user_data,
        )
        assert isinstance(result, bytes)

    def test_build_mcs_send_data_request_large_data(self) -> None:
        """Test building MCS send data request with large data."""
        user_data = b"\x00" * 5000
        result = build_mcs_send_data_request(
            user_id=1001,
            channel_id=1003,
            user_data=user_data,
        )
        assert isinstance(result, bytes)

    def test_build_mcs_send_data_request_too_large(self) -> None:
        """Test building MCS send data request with too large data."""
        user_data = b"\x00" * 20000  # > 16383 bytes
        with pytest.raises(ValueError, match="too large"):
            build_mcs_send_data_request(
                user_id=1001,
                channel_id=1003,
                user_data=user_data,
            )

    def test_build_mcs_send_data_request_custom_priority(self) -> None:
        """Test building MCS send data request with custom priority."""
        result = build_mcs_send_data_request(
            user_id=1001,
            channel_id=1003,
            data_priority=2,
            user_data=b"\x00" * 10,
        )
        assert isinstance(result, bytes)
        # Data priority should be in lower 2 bits
        assert (result[0] & 0x03) == 2

    def test_build_mcs_send_data_request_empty_data(self) -> None:
        """Test building MCS send data request with empty data."""
        result = build_mcs_send_data_request(
            user_id=1001,
            channel_id=1003,
            user_data=b"",
        )
        assert isinstance(result, bytes)
        # Should still have header bytes
        assert len(result) >= 6


class TestMcsErectDomainRequest:
    """Tests for MCS Erect Domain Request building."""

    def test_build_mcs_erect_domain_request_custom_values(self) -> None:
        """Test building erect domain request with custom values."""
        result = build_mcs_erect_domain_request(sub_height=5, sub_interval=10)
        assert len(result) > 0
        assert result[0] == MCS_TYPE_ERECT_DOMAIN_REQUEST


class TestMcsChannelJoinRequest:
    """Tests for MCS Channel Join Request building."""

    def test_build_mcs_channel_join_request_various_channels(self) -> None:
        """Test building channel join request for various channels."""
        for channel_id in [1003, 1004, 1005, 1006]:
            result = build_mcs_channel_join_request(user_id=1001, channel_id=channel_id)
            assert result[0] == MCS_TYPE_CHANNEL_JOIN_REQUEST



#!/usr/bin/env python3
"""
Graphics Test with Display class - Raw frames + Live ffmpeg encoding.

Uses the new Display class which:
- Stores raw RGB frames (no PNG encoding overhead)
- Runs ffmpeg in background for live video encoding
- Maintains 100MB video output buffer with auto-eviction

Run with:
    export RDP_HOST=... RDP_USER=... RDP_PASS=... DURATION=30
    python example_display_profiled.py

Then visualize with:
    snakeviz sessions/display_XXXX/profile.prof
"""

import asyncio
import cProfile
import os
import pstats
import sys
import time
from datetime import datetime

from dotenv import load_dotenv

load_dotenv()

# Add src to path
sys.path.insert(0, os.path.join(os.path.dirname(__file__), "src"))

from simple_rdp import RDPClient  # noqa: E402
from simple_rdp.display import Display  # noqa: E402


class PerformanceTracker:
    """Track performance metrics during capture."""

    def __init__(self):
        self.screenshot_times: list[float] = []
        self.add_frame_times: list[float] = []
        self.start_time: float = 0
        self.end_time: float = 0

    @property
    def duration(self) -> float:
        return self.end_time - self.start_time if self.end_time else 0

    def avg_ms(self, times: list[float]) -> float:
        return sum(times) / len(times) * 1000 if times else 0

    def print_stats(self):
        print(f"\n{'=' * 60}")
        print("           TIMING BREAKDOWN")
        print(f"{'=' * 60}")

        if self.screenshot_times:
            print(f"\n📷 SCREENSHOT ({len(self.screenshot_times)} samples):")
            print(f"   Average:  {self.avg_ms(self.screenshot_times):.2f} ms")
            print(f"   Min:      {min(self.screenshot_times) * 1000:.2f} ms")
            print(f"   Max:      {max(self.screenshot_times) * 1000:.2f} ms")
            print(f"   Total:    {sum(self.screenshot_times):.2f} s")

        if self.add_frame_times:
            print(f"\n🖼️  ADD FRAME (raw bytes + ffmpeg write) ({len(self.add_frame_times)} samples):")
            print(f"   Average:  {self.avg_ms(self.add_frame_times):.2f} ms")
            print(f"   Min:      {min(self.add_frame_times) * 1000:.2f} ms")
            print(f"   Max:      {max(self.add_frame_times) * 1000:.2f} ms")
            print(f"   Total:    {sum(self.add_frame_times):.2f} s")

        if self.screenshot_times and self.add_frame_times:
            total_ss = sum(self.screenshot_times)
            total_add = sum(self.add_frame_times)
            total = total_ss + total_add
            print("\n⏱️  TIME DISTRIBUTION:")
            print(f"   Screenshot:  {total_ss / total * 100:.1f}% ({total_ss:.2f}s)")
            print(f"   Add frame:   {total_add / total * 100:.1f}% ({total_add:.2f}s)")


async def capture_loop(
    client: RDPClient,
    display: Display,
    tracker: PerformanceTracker,
    stop_event: asyncio.Event,
    target_fps: int = 30,
):
    """
    Capture frames at target FPS using raw frame storage.

    This is MUCH faster than PNG encoding because:
    - image.tobytes() takes ~1-2ms vs PNG encoding ~150-200ms
    - ffmpeg encoding happens in separate process
    """
    target_interval = 1.0 / target_fps
    frame_count = 0

    while not stop_event.is_set():
        loop_start = time.perf_counter()

        try:
            # Get screenshot (just a buffer copy - very fast)
            t0 = time.perf_counter()
            screenshot = await client.screenshot()
            t1 = time.perf_counter()
            tracker.screenshot_times.append(t1 - t0)

            # Add to display (converts to raw bytes + sends to ffmpeg)
            t2 = time.perf_counter()
            await display.add_frame(screenshot)
            t3 = time.perf_counter()
            tracker.add_frame_times.append(t3 - t2)

            frame_count += 1

            if frame_count % 100 == 0:
                avg_ss = tracker.avg_ms(tracker.screenshot_times[-100:])
                avg_add = tracker.avg_ms(tracker.add_frame_times[-100:])
                print(
                    f"    [Capture] {frame_count} frames | "
                    f"screenshot: {avg_ss:.1f}ms, add_frame: {avg_add:.1f}ms | "
                    f"video buf: {display.video_buffer_size_mb:.1f}MB"
                )

        except Exception as e:
            print(f"    [Capture Error] {e}")

        # Sleep for remainder of interval
        elapsed = time.perf_counter() - loop_start
        sleep_time = max(0, target_interval - elapsed)
        await asyncio.sleep(sleep_time)


async def ui_actions(client: RDPClient, duration_seconds: int = 30):
    """
    Perform UI actions to generate screen changes.
    Faster action interval since capture is now faster.
    """
    print(f"\n=== Starting {duration_seconds}s of UI actions ===\n")

    action_interval = 0.3  # Faster: 0.3s between actions

    actions = [
        (
            "Open Task Manager",
            [("ctrl", True), ("shift", True), ("escape", True)],
            [("escape", False), ("shift", False), ("ctrl", False)],
        ),
        ("Close Window", [("alt", True), ("f4", True)], [("f4", False), ("alt", False)]),
        ("Open Search", [("win", True)], [("win", False)]),
        ("Close Search", [("escape", True)], [("escape", False)]),
        ("Open Run", [("win", True), ("r", True)], [("r", False), ("win", False)]),
        ("Close Run", [("escape", True)], [("escape", False)]),
        ("Open Settings", [("win", True), ("i", True)], [("i", False), ("win", False)]),
        ("Close Settings", [("alt", True), ("f4", True)], [("f4", False), ("alt", False)]),
        ("Open Explorer", [("win", True), ("e", True)], [("e", False), ("win", False)]),
        ("Close Explorer", [("alt", True), ("f4", True)], [("f4", False), ("alt", False)]),
        ("Show Desktop", [("win", True), ("d", True)], [("d", False), ("win", False)]),
        ("Restore Desktop", [("win", True), ("d", True)], [("d", False), ("win", False)]),
    ]

    scancodes = {
        "escape": 0x01,
        "f4": 0x3E,
        "ctrl": 0x1D,
        "shift": 0x2A,
        "alt": 0x38,
        "win": 0x5B,
        "r": 0x13,
        "i": 0x17,
        "e": 0x12,
        "d": 0x20,
    }

    start_time = time.time()
    action_count = 0
    action_idx = 0

    while (time.time() - start_time) < duration_seconds:
        action_name, press_keys, release_keys = actions[action_idx]
        action_idx = (action_idx + 1) % len(actions)
        action_count += 1

        elapsed = time.time() - start_time
        print(f"  [{elapsed:5.1f}s] Action {action_count}: {action_name}")

        try:
            for key, _ in press_keys:
                scancode = scancodes.get(key, 0)
                await client.send_key(scancode, is_press=True, is_release=False)
                await asyncio.sleep(0.015)

            await asyncio.sleep(0.03)

            for key, _ in release_keys:
                scancode = scancodes.get(key, 0)
                await client.send_key(scancode, is_press=False, is_release=True)
                await asyncio.sleep(0.015)

        except Exception as e:
            print(f"    [Action Error] {e}")

        await asyncio.sleep(action_interval)

    print(f"\n=== Completed {action_count} actions in {duration_seconds}s ===\n")


async def run_test(session_dir: str, duration: int, target_fps: int):
    """Main test logic."""
    host = os.environ.get("RDP_HOST", "")
    username = os.environ.get("RDP_USER", "")
    password = os.environ.get("RDP_PASS", "")

    if not host or not username or not password:
        print("ERROR: Set RDP_HOST, RDP_USER, RDP_PASS environment variables (or use .env file)")
        return None, None

    # Create display with 100MB video buffer limit
    display = Display(
        width=1920,
        height=1080,
        fps=target_fps,
        max_video_buffer_mb=100,
        max_raw_frames=500,  # Keep last 500 raw frames
    )

    tracker = PerformanceTracker()
    stop_capture = asyncio.Event()

    print(f"\nConnecting to {host} as {username}...")

    client = RDPClient(host, username=username, password=password, width=1920, height=1080, show_wallpaper=True)

    try:
        await client.connect()
        print("Connected! Waiting for initial screen...")
        await asyncio.sleep(2)

        # Start ffmpeg streaming
        print(f"\nStarting ffmpeg live streaming at {target_fps} FPS...")
        await display.start_streaming()

        # Start capture
        print(f"Starting {target_fps} FPS capture with raw frames...")
        tracker.start_time = time.perf_counter()
        capture_task = asyncio.create_task(capture_loop(client, display, tracker, stop_capture, target_fps))

        await asyncio.sleep(0.5)

        # Run UI actions
        await ui_actions(client, duration_seconds=duration)

        # Stop capture
        print("\nStopping capture...")
        stop_capture.set()
        await capture_task
        tracker.end_time = time.perf_counter()

        # Stop streaming
        await display.stop_streaming()

        print(f"\nCapture complete: {display.frame_count} frames in {tracker.duration:.1f}s")
        print(f"Actual FPS: {display.frame_count / tracker.duration:.2f}")

    except Exception as e:
        print(f"Error: {e}")
        import traceback

        traceback.print_exc()
        stop_capture.set()
        tracker.end_time = time.perf_counter()
    finally:
        await display.stop_streaming()
        await client.disconnect()

    return display, tracker


def main():
    duration = int(os.environ.get("DURATION", "30"))
    target_fps = int(os.environ.get("TARGET_FPS", "30"))  # Higher FPS now possible!

    # Create session directory
    session_id = datetime.now().strftime("%Y%m%d_%H%M%S")
    session_dir = os.path.join(os.path.dirname(__file__), "sessions", f"display_{session_id}")
    os.makedirs(session_dir, exist_ok=True)

    profile_path = os.path.join(session_dir, "profile.prof")
    video_path = os.path.join(session_dir, "recording.ts")

    print("=" * 60)
    print("     DISPLAY CLASS TEST with cProfile + SNAKEVIZ")
    print("=" * 60)
    print(f"\nSession directory: {session_dir}")
    print(f"Test duration: {duration} seconds")
    print(f"Target FPS: {target_fps}")
    print(f"Profile output: {profile_path}")

    # Create profiler
    profiler = cProfile.Profile()

    # Run the test with profiling
    print("\n[Starting profiled test run...]")
    profiler.enable()

    display, tracker = asyncio.run(run_test(session_dir, duration, target_fps))

    profiler.disable()
    print("[Profiling complete]")

    # Save profile data
    profiler.dump_stats(profile_path)
    print(f"\nProfile saved to: {profile_path}")

    if display and tracker:
        # Print timing stats
        tracker.print_stats()

        # Print display stats
        display.print_stats()

        # Save video from buffer
        if display.video_buffer_size_mb > 0:
            print(f"\nSaving video buffer to {video_path}...")
            asyncio.run(display.save_video(video_path))
            video_size = os.path.getsize(video_path) / (1024 * 1024)
            print(f"Video saved: {video_size:.2f} MB")

        # Also save from raw frames as MP4
        raw_video_path = os.path.join(session_dir, "recording_raw.mp4")
        print(f"\nEncoding raw frames to {raw_video_path}...")
        asyncio.run(display.save_raw_frames_as_video(raw_video_path, fps=target_fps))
        if os.path.exists(raw_video_path):
            raw_video_size = os.path.getsize(raw_video_path) / (1024 * 1024)
            print(f"Raw frames video: {raw_video_size:.2f} MB")

        # Performance summary
        print("\n" + "=" * 60)
        print("              PERFORMANCE SUMMARY")
        print("=" * 60)
        print(f"\n📊 Total frames:     {display.frame_count}")
        print(f"   Duration:         {tracker.duration:.1f}s")
        print(f"   Actual FPS:       {display.frame_count / tracker.duration:.2f}")
        print(f"   Target FPS:       {target_fps}")

        raw_mb = display.raw_frame_count * 1920 * 1080 * 3 / (1024 * 1024)
        print(f"\n💾 Raw frames kept:  {display.raw_frame_count} ({raw_mb:.1f} MB)")
        print(f"   Video buffer:     {display.video_buffer_size_mb:.2f} MB")

    # Print top 20 functions
    print("\n" + "=" * 60)
    print("         TOP 20 FUNCTIONS BY CUMULATIVE TIME")
    print("=" * 60 + "\n")
    stats = pstats.Stats(profiler)
    stats.sort_stats("cumulative")
    stats.print_stats(20)

    print("\n" + "=" * 60)
    print("         VIEW FULL PROFILE WITH SNAKEVIZ")
    print("=" * 60)
    print("\nRun this command to visualize:")
    print(f"  snakeviz {profile_path}")
    print()

    return profile_path


if __name__ == "__main__":
    main()



"""Tests for RDP Client."""

import tempfile
from unittest.mock import MagicMock
from unittest.mock import patch

import pytest

from simple_rdp.client import IO_CHANNEL_ID
from simple_rdp.client import MCS_GLOBAL_CHANNEL_ID
from simple_rdp.client import RDPClient


class TestRDPClient:
    """Tests for RDPClient class."""

    def test_client_initialization(self):
        """Test client can be initialized with host."""
        client = RDPClient(host="localhost")
        assert client.is_connected is False
        assert client.host == "localhost"
        assert client.port == 3389

    def test_client_custom_port(self):
        """Test client with custom port."""
        client = RDPClient(host="localhost", port=3390)
        assert client.port == 3390

    def test_client_full_params(self):
        """Test client with all parameters."""
        client = RDPClient(
            host="server.example.com",
            port=3389,
            username="user",
            password="pass",
            domain="DOMAIN",
        )
        assert client.host == "server.example.com"

    def test_client_dimensions(self):
        """Test client dimension properties."""
        client = RDPClient(host="localhost", width=1280, height=720)
        assert client.width == 1280
        assert client.height == 720

    def test_client_default_dimensions(self):
        """Test client default dimensions."""
        client = RDPClient(host="localhost")
        assert client.width == 1920
        assert client.height == 1080

    def test_client_connection_properties_initial(self):
        """Test client connection properties are empty initially."""
        client = RDPClient(host="localhost")
        assert client.connection_properties == {}

    @pytest.mark.asyncio
    async def test_disconnect_sets_connected_false(self):
        """Test that disconnect sets connected state to False."""
        client = RDPClient(host="localhost")
        await client.disconnect()
        assert client.is_connected is False

    @pytest.mark.asyncio
    async def test_connect_fails_on_invalid_host(self):
        """Test that connect raises ConnectionError on invalid host."""
        client = RDPClient(host="invalid.host.that.does.not.exist.local", port=3389)
        with pytest.raises(ConnectionError):
            await client.connect()


class TestClientConstants:
    """Tests for client constants."""

    def test_io_channel_id(self):
        """Test IO_CHANNEL_ID value."""
        assert IO_CHANNEL_ID == 1003

    def test_mcs_global_channel_id(self):
        """Test MCS_GLOBAL_CHANNEL_ID value."""
        assert MCS_GLOBAL_CHANNEL_ID == 1003


class TestClientProperties:
    """Tests for client property accessors."""

    def test_host_property(self):
        """Test host property."""
        client = RDPClient(host="test-server")
        assert client.host == "test-server"

    def test_port_property(self):
        """Test port property."""
        client = RDPClient(host="localhost", port=13389)
        assert client.port == 13389

    def test_is_connected_false_by_default(self):
        """Test is_connected is False by default."""
        client = RDPClient(host="localhost")
        assert client.is_connected is False

    def test_width_property(self):
        """Test width property."""
        client = RDPClient(host="localhost", width=800)
        assert client.width == 800

    def test_height_property(self):
        """Test height property."""
        client = RDPClient(host="localhost", height=600)
        assert client.height == 600


class TestClientContextManager:
    """Tests for client context manager support."""

    @pytest.mark.asyncio
    async def test_client_context_manager_no_connect(self):
        """Test client can be used as context manager without connecting."""
        # Client.__aenter__ and __aexit__ should work even if connect fails
        client = RDPClient(host="localhost")
        # Just test that the client can be instantiated
        assert client is not None


class TestClientMultipleInstances:
    """Tests for multiple client instances."""

    def test_multiple_clients_independent(self):
        """Test multiple clients are independent."""
        client1 = RDPClient(host="host1", port=3389)
        client2 = RDPClient(host="host2", port=3390)

        assert client1.host != client2.host
        assert client1.port != client2.port

    def test_client_default_values(self):
        """Test client default values."""
        client = RDPClient(host="localhost")
        assert client.width == 1920
        assert client.height == 1080
        # Check it starts disconnected
        assert not client.is_connected


class TestClientReaderWriter:
    """Tests for client reader/writer property access."""

    def test_reader_raises_when_not_connected(self):
        """Test _reader property raises when not connected."""
        client = RDPClient(host="localhost")
        # The _reader property should raise ConnectionError
        with pytest.raises(ConnectionError, match="Not connected"):
            _ = client._reader

    def test_writer_raises_when_not_connected(self):
        """Test _writer property raises when not connected."""
        client = RDPClient(host="localhost")
        # The _writer property should raise ConnectionError
        with pytest.raises(ConnectionError, match="Not connected"):
            _ = client._writer


class TestClientInternalState:
    """Tests for client internal state."""

    def test_initial_user_id(self):
        """Test initial user_id is 0."""
        client = RDPClient(host="localhost")
        assert client._user_id == 0

    def test_initial_io_channel_id(self):
        """Test initial _io_channel_id."""
        client = RDPClient(host="localhost")
        assert client._io_channel_id == IO_CHANNEL_ID

    def test_initial_channel_ids_empty(self):
        """Test initial _channel_ids is empty."""
        client = RDPClient(host="localhost")
        assert client._channel_ids == []

    def test_initial_share_id(self):
        """Test initial _share_id is 0."""
        client = RDPClient(host="localhost")
        assert client._share_id == 0

    def test_initial_screen_buffer_none(self):
        """Test initial display screen_buffer is None."""
        client = RDPClient(host="localhost")
        assert client._display._screen_buffer is None

    def test_initial_fragment_buffer_empty(self):
        """Test initial fragment buffer is empty."""
        client = RDPClient(host="localhost")
        assert len(client._fragment_buffer) == 0
        assert client._fragment_type == 0

    def test_initial_running_false(self):
        """Test _running is False initially."""
        client = RDPClient(host="localhost")
        assert client._running is False

    def test_initial_receive_task_none(self):
        """Test _receive_task is None initially."""
        client = RDPClient(host="localhost")
        assert client._receive_task is None


class TestClientScreenCapture:
    """Tests for client screen capture methods."""

    @pytest.mark.asyncio
    async def test_screenshot_returns_blank_when_not_connected(self):
        """Test screenshot returns blank image when not connected."""
        client = RDPClient(host="localhost", width=100, height=100)
        # Display screen_buffer is None, should return blank image
        img = await client.screenshot()
        assert img.size == (100, 100)
        # Should be all black
        assert img.getpixel((0, 0)) == (0, 0, 0)

    @pytest.mark.asyncio
    async def test_screenshot_with_screen_buffer(self):
        """Test screenshot with initialized screen buffer."""
        from PIL import Image

        client = RDPClient(host="localhost", width=100, height=100)
        # Initialize and set screen buffer via display
        client._display.initialize_screen()
        client._display._screen_buffer = Image.new("RGB", (100, 100), color=(255, 0, 0))

        img = await client.screenshot()
        assert img.size == (100, 100)
        # Should be red
        assert img.getpixel((50, 50)) == (255, 0, 0)
        # Should be a copy, not the original
        assert img is not client._display._screen_buffer


class TestClientColorDepth:
    """Tests for client color depth settings."""

    def test_default_color_depth(self):
        """Test default color depth is 32."""
        client = RDPClient(host="localhost")
        assert client._color_depth == 32

    def test_custom_color_depth(self):
        """Test custom color depth."""
        client = RDPClient(host="localhost", color_depth=24)
        assert client._color_depth == 24


class TestClientWallpaper:
    """Tests for client wallpaper settings."""

    def test_default_show_wallpaper_false(self):
        """Test show_wallpaper is False by default."""
        client = RDPClient(host="localhost")
        assert client._show_wallpaper is False

    def test_show_wallpaper_true(self):
        """Test show_wallpaper can be True."""
        client = RDPClient(host="localhost", show_wallpaper=True)
        assert client._show_wallpaper is True


class TestClientAuth:
    """Tests for client authentication settings."""

    def test_username_none_by_default(self):
        """Test username is None by default."""
        client = RDPClient(host="localhost")
        assert client._username is None

    def test_password_none_by_default(self):
        """Test password is None by default."""
        client = RDPClient(host="localhost")
        assert client._password is None

    def test_domain_none_by_default(self):
        """Test domain is None by default."""
        client = RDPClient(host="localhost")
        assert client._domain is None

    def test_auth_credentials_set(self):
        """Test authentication credentials are set correctly."""
        client = RDPClient(
            host="localhost",
            username="testuser",
            password="testpass",
            domain="TESTDOMAIN",
        )
        assert client._username == "testuser"
        assert client._password == "testpass"
        assert client._domain == "TESTDOMAIN"


class TestClientDisconnect:
    """Tests for client disconnect behavior."""

    @pytest.mark.asyncio
    async def test_disconnect_clears_writers(self):
        """Test disconnect clears reader/writer."""
        client = RDPClient(host="localhost")
        await client.disconnect()
        assert client._tcp_reader is None
        assert client._tcp_writer is None

    @pytest.mark.asyncio
    async def test_disconnect_sets_running_false(self):
        """Test disconnect sets _running to False."""
        client = RDPClient(host="localhost")
        client._running = True  # Simulate running
        await client.disconnect()
        assert client._running is False

    @pytest.mark.asyncio
    async def test_disconnect_multiple_times(self):
        """Test disconnect can be called multiple times safely."""
        client = RDPClient(host="localhost")
        await client.disconnect()
        await client.disconnect()
        await client.disconnect()
        assert client.is_connected is False


class TestClientDisplay:
    """Tests for client Display integration."""

    def test_display_property(self):
        """Test display property returns Display instance."""
        from simple_rdp.display import Display

        client = RDPClient(host="localhost")
        assert isinstance(client.display, Display)

    def test_display_dimensions_match(self):
        """Test display dimensions match client dimensions."""
        client = RDPClient(host="localhost", width=1280, height=720)
        assert client.display.width == 1280
        assert client.display.height == 720

    def test_display_default_fps(self):
        """Test display has default 30 fps."""
        client = RDPClient(host="localhost")
        assert client.display.fps == 30

    def test_is_streaming_false_by_default(self):
        """Test is_streaming is False by default."""
        client = RDPClient(host="localhost")
        assert client.is_streaming is False

    @pytest.mark.asyncio
    @patch("subprocess.Popen")
    async def test_start_streaming(self, mock_popen):
        """Test start_streaming sets streaming flag."""
        mock_popen.return_value = MagicMock()
        client = RDPClient(host="localhost")
        await client.start_streaming()
        assert client.is_streaming is True
        await client.stop_streaming()

    @pytest.mark.asyncio
    @patch("subprocess.Popen")
    async def test_stop_streaming(self, mock_popen):
        """Test stop_streaming clears streaming flag."""
        mock_popen.return_value = MagicMock()
        client = RDPClient(host="localhost")
        await client.start_streaming()
        await client.stop_streaming()
        assert client.is_streaming is False

    @pytest.mark.asyncio
    @patch("subprocess.Popen")
    async def test_start_file_recording(self, mock_popen):
        """Test start_file_recording starts both streaming and file recording."""
        mock_popen.return_value = MagicMock()
        client = RDPClient(host="localhost")
        # Use tempfile for cross-platform temp path
        with tempfile.NamedTemporaryFile(suffix=".ts", delete=False) as f:
            temp_path = f.name
        await client.start_file_recording(temp_path)
        assert client.is_streaming is True
        assert client.is_file_recording is True
        await client.stop_streaming()

    def test_get_recording_stats(self):
        """Test get_recording_stats returns stats dict."""
        client = RDPClient(host="localhost")
        stats = client.get_recording_stats()
        assert "frames_received" in stats
        assert "frames_encoded" in stats
        assert "encoding_errors" in stats

    @patch("subprocess.run")
    def test_transcode_success(self, mock_run):
        """Test transcode returns True on success."""
        mock_run.return_value = MagicMock(returncode=0)
        result = RDPClient.transcode("input.ts", "output.mp4")
        assert result is True
        mock_run.assert_called_once()
        # Verify ffmpeg was called with correct arguments
        call_args = mock_run.call_args[0][0]
        assert "ffmpeg" in call_args
        assert "input.ts" in call_args
        assert "output.mp4" in call_args
        assert "-c" in call_args
        assert "copy" in call_args

    @patch("subprocess.run")
    def test_transcode_failure(self, mock_run):
        """Test transcode returns False on ffmpeg error."""
        import subprocess

        mock_run.side_effect = subprocess.CalledProcessError(1, "ffmpeg", stderr="error")
        result = RDPClient.transcode("input.ts", "output.mp4")
        assert result is False

    @patch("subprocess.run")
    def test_transcode_ffmpeg_not_found(self, mock_run):
        """Test transcode returns False when ffmpeg is not installed."""
        mock_run.side_effect = FileNotFoundError()
        result = RDPClient.transcode("input.ts", "output.mp4")
        assert result is False

    @pytest.mark.asyncio
    @patch("subprocess.Popen")
    async def test_disconnect_stops_streaming(self, mock_popen):
        """Test disconnect stops streaming if active."""
        mock_popen.return_value = MagicMock()
        client = RDPClient(host="localhost")
        await client.start_streaming()
        assert client.is_streaming is True
        await client.disconnect()
        assert client.is_streaming is False
